"""Module for VeSync Fans (Not Purifiers or Humidifiers)."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from pyvesync.base_devices import VeSyncFanBase
from pyvesync.const import ConnectionStatus, DeviceStatus
from pyvesync.models.bypass_models import TimerModels
from pyvesync.models.fan_models import PedestalFanResult, TowerFanResult
from pyvesync.utils.device_mixins import BypassV2Mixin, process_bypassv2_result
from pyvesync.utils.helpers import Helpers, Timer

if TYPE_CHECKING:
    from pyvesync import VeSync
    from pyvesync.device_map import FanMap
    from pyvesync.models.vesync_models import ResponseDeviceDetailsModel

logger = logging.getLogger(__name__)


class VeSyncTowerFan(BypassV2Mixin, VeSyncFanBase):
    """Levoit Tower Fan Device Class."""

    __slots__ = ()

    def __init__(
        self,
        details: ResponseDeviceDetailsModel,
        manager: VeSync,
        feature_map: FanMap,
    ) -> None:
        """Initialize the VeSync Base API V2 Fan Class."""
        super().__init__(details, manager, feature_map)

    def _set_fan_state(self, res: TowerFanResult) -> None:
        """Set fan state attributes from API response."""
        self.state.connection_status = ConnectionStatus.ONLINE
        self.state.device_status = DeviceStatus.from_int(res.powerSwitch)
        self.state.mode = self._reverse_modes.get(res.workMode)
        self.state.fan_level = res.fanSpeedLevel
        self.state.fan_set_level = res.manualSpeedLevel
        self.state.temperature = res.temperature

        self.state.mute_status = DeviceStatus.from_int(res.muteState)
        self.state.mute_set_status = DeviceStatus.from_int(res.muteSwitch)
        self.state.oscillation_status = DeviceStatus.from_int(res.oscillationState)
        self.state.oscillation_set_status = DeviceStatus.from_int(res.oscillationSwitch)
        self.state.display_status = DeviceStatus.from_int(res.screenState)
        self.state.display_set_status = DeviceStatus.from_int(res.screenSwitch)
        self.state.displaying_type = DeviceStatus.from_int(res.displayingType)

        if res.timerRemain is not None and res.timerRemain > 0:
            if self.state.device_status == DeviceStatus.ON:
                self.state.timer = Timer(res.timerRemain, action='off')
            else:
                self.state.timer = Timer(res.timerRemain, action='on')
        if res.sleepPreference is None:
            return
        self.state.sleep_preference_type = res.sleepPreference.sleepPreferenceType
        self.state.sleep_fallasleep_remain = DeviceStatus.from_int(
            res.sleepPreference.fallAsleepRemain
        )
        self.state.sleep_oscillation_switch = DeviceStatus.from_int(
            res.sleepPreference.oscillationSwitch
        )
        self.state.sleep_change_fan_level = DeviceStatus.from_int(
            res.sleepPreference.autoChangeFanLevelSwitch
        )

    async def get_details(self) -> None:
        r_dict = await self.call_bypassv2_api('getTowerFanStatus')
        result = process_bypassv2_result(
            self, logger, 'get_details', r_dict, TowerFanResult
        )
        if result is None:
            return
        self._set_fan_state(result)

    async def set_fan_speed(self, speed: int | None = None) -> bool:
        if speed is None:
            new_speed = Helpers.bump_level(self.state.fan_level, self.fan_levels)
        else:
            new_speed = speed

        if new_speed not in self.fan_levels:
            logger.warning('Invalid fan speed level used - %s', speed)
            return False

        payload_data = {'manualSpeedLevel': speed, 'levelType': 'wind', 'levelIdx': 0}
        r_dict = await self.call_bypassv2_api('setLevel', payload_data)
        r = Helpers.process_dev_response(logger, 'set_fan_speed', self, r_dict)
        if r is None:
            return False

        self.state.fan_level = speed
        self.state.fan_set_level = speed
        self.state.connection_status = ConnectionStatus.ONLINE
        return True

    async def set_mode(self, mode: str) -> bool:
        if mode not in self.modes:
            logger.warning('Invalid purifier mode used - %s', mode)
            return False

        payload_data = {'workMode': self.modes[mode]}
        r_dict = await self.call_bypassv2_api('setTowerFanMode', payload_data)
        r = Helpers.process_dev_response(logger, 'set_mode', self, r_dict)
        if r is None:
            return False

        self.state.mode = mode
        self.state.connection_status = ConnectionStatus.ONLINE
        return True

    async def toggle_switch(self, toggle: bool | None = None) -> bool:
        if toggle is None:
            toggle = self.state.device_status == DeviceStatus.OFF
        payload_data = {
            'powerSwitch': int(toggle),
            'switchIdx': 0,
        }
        r_dict = await self.call_bypassv2_api('setSwitch', payload_data)
        r = Helpers.process_dev_response(logger, 'toggle_switch', self, r_dict)
        if r is None:
            return False

        self.state.device_status = DeviceStatus.from_bool(toggle)
        self.state.connection_status = ConnectionStatus.ONLINE
        return True

    async def toggle_oscillation(self, toggle: bool) -> bool:
        payload_data = {'oscillationSwitch': int(toggle)}
        r_dict = await self.call_bypassv2_api('setOscillationSwitch', payload_data)
        r = Helpers.process_dev_response(logger, 'toggle_oscillation', self, r_dict)
        if r is None:
            return False

        self.state.oscillation_status = DeviceStatus.from_bool(toggle)
        self.state.connection_status = ConnectionStatus.ONLINE
        return True

    async def toggle_mute(self, toggle: bool) -> bool:
        payload_data = {'muteSwitch': int(toggle)}
        r_dict = await self.call_bypassv2_api('setMuteSwitch', payload_data)
        r = Helpers.process_dev_response(logger, 'toggle_mute', self, r_dict)
        if r is None:
            return False

        self.state.mute_status = DeviceStatus.from_bool(toggle)
        self.state.connection_status = ConnectionStatus.ONLINE
        return True

    async def toggle_display(self, toggle: bool) -> bool:
        payload_data = {'screenSwitch': int(toggle)}
        r_dict = await self.call_bypassv2_api('setDisplay', payload_data)
        r = Helpers.process_dev_response(logger, 'toggle_display', self, r_dict)
        if r is None:
            return False

        self.state.display_status = DeviceStatus.from_bool(toggle)
        self.state.connection_status = ConnectionStatus.ONLINE
        return True

    async def toggle_displaying_type(self, toggle: bool) -> bool:
        """Set displaying type on/off - Unknown functionality."""
        payload_data = {'displayingType': int(toggle)}
        r_dict = await self.call_bypassv2_api('setDisplayingType', payload_data)
        r = Helpers.process_dev_response(logger, 'toggle_displaying_type', self, r_dict)
        if r is None:
            return False

        self.state.displaying_type = DeviceStatus.from_bool(toggle)
        self.state.connection_status = ConnectionStatus.ONLINE
        return True

    async def get_timer(self) -> None:
        r_dict = await self.call_bypassv2_api('getTimer', {})
        r_model = process_bypassv2_result(
            self, logger, 'get_timer', r_dict, TimerModels.ResultV2GetTimer
        )
        if not r_model:
            logger.debug('No timer found - run get_timer()')
            return
        if r_model.timers is None or len(r_model.timers) == 0:
            logger.debug('No timer found - run get_timer()')
            return
        timer = r_model.timers[0]
        if not isinstance(timer, TimerModels.TimerItemV2):
            logger.warning('Invalid timer found for %s', self.device_name)
            return
        self.state.timer = Timer(timer.remain, timer.action, timer.id)

    async def set_timer(self, duration: int, action: str | None = None) -> bool:
        if action is None:
            action = (
                DeviceStatus.OFF
                if self.state.device_status == DeviceStatus.ON
                else DeviceStatus.ON
            )
        if action not in [DeviceStatus.ON, DeviceStatus.OFF]:
            logger.warning('Invalid action used - %s', action)
            return False
        payload_data = {
            'action': action,
            'total': duration,
        }
        r_dict = await self.call_bypassv2_api('setTimer', payload_data)
        r_model = process_bypassv2_result(
            self, logger, 'set_timer', r_dict, TimerModels.ResultV2SetTimer
        )
        if r_model is None:
            return False
        self.state.timer = Timer(duration, action, r_model.id)
        return True

    async def clear_timer(self) -> bool:
        if self.state.timer is None:
            logger.debug('No timer found for - run get_timer()')
            return False
        payload_data = {
            'id': self.state.timer.id,
        }
        r_dict = await self.call_bypassv2_api('clearTimer', payload_data)
        result = Helpers.process_dev_response(logger, 'clear_timer', self, r_dict)
        if result is None:
            return False
        self.state.timer = None
        return True


class VeSyncPedestalFan(BypassV2Mixin, VeSyncFanBase):
    """Levoit Pedestal Fan Device Class."""

    __slots__ = ()

    def __init__(
        self,
        details: ResponseDeviceDetailsModel,
        manager: VeSync,
        feature_map: FanMap,
    ) -> None:
        """Initialize the VeSync Base API V2 Fan Class."""
        super().__init__(details, manager, feature_map)

    def _set_fan_state(self, res: PedestalFanResult) -> None:
        """Set the fan state from the result."""
        self.state.device_status = DeviceStatus.from_int(res.powerSwitch)
        self.state.mode = self._reverse_modes.get(res.workMode)
        self.state.fan_level = res.fanSpeedLevel
        self.state.temperature = (res.temperature / 10) if res.temperature else None
        self.state.mute_set_status = DeviceStatus.from_int(res.muteSwitch)
        self.state.display_set_status = DeviceStatus.from_int(res.screenSwitch)
        self.state.display_status = DeviceStatus.from_int(res.screenState)
        self.state.vertical_oscillation_status = DeviceStatus.from_int(
            res.verticalOscillationState
        )
        self.state.horizontal_oscillation_status = DeviceStatus.from_int(
            res.horizontalOscillationState
        )
        self.state.child_lock = bool(res.childLock)
        if res.sleepPreference is not None:
            self.state.sleep_change_fan_level = DeviceStatus.from_int(
                res.sleepPreference.initFanSpeedLevel
            )
            self.state.sleep_fallasleep_remain = DeviceStatus.from_int(
                res.sleepPreference.fallAsleepRemain
            )
            self.state.sleep_oscillation_switch = DeviceStatus.from_int(
                res.sleepPreference.oscillationState
            )
            self.state.sleep_preference_type = res.sleepPreference.sleepPreferenceType

    async def get_details(self) -> None:
        r_dict = await self.call_bypassv2_api('getFanStatus')
        result = process_bypassv2_result(
            self, logger, 'get_details', r_dict, PedestalFanResult
        )
        if result is None:
            return
        self._set_fan_state(result)

    async def set_mode(self, mode: str) -> bool:
        if mode not in self.modes:
            logger.warning('Invalid fan mode used - %s', mode)
            return False
        data = {'workMode': self.modes[mode]}
        r_dict = await self.call_bypassv2_api('setFanMode', data)
        r = Helpers.process_dev_response(logger, 'set_mode', self, r_dict)
        if r is None:
            return False
        self.state.mode = mode
        self.state.connection_status = ConnectionStatus.ONLINE
        return True

    async def set_fan_speed(self, speed: int | None = None) -> bool:
        if speed is None:
            new_speed = Helpers.bump_level(self.state.fan_level, self.fan_levels)
        else:
            new_speed = speed
        if new_speed not in self.fan_levels:
            logger.warning('Invalid fan speed level used - %s', speed)
            return False
        data = {'levelIdx': 0, 'levelType': 'wind', 'manualSpeedLevel': speed}
        r_dict = await self.call_bypassv2_api('setLevel', data)
        r = Helpers.process_dev_response(logger, 'set_fan_speed', self, r_dict)
        if r is None:
            return False
        self.state.fan_level = speed
        self.state.connection_status = ConnectionStatus.ONLINE
        return True

    async def toggle_switch(self, toggle: bool | None = None) -> bool:
        if toggle is None:
            toggle = self.state.device_status == DeviceStatus.OFF
        data = {'powerSwitch': int(toggle), 'switchIdx': 0}
        r_dict = await self.call_bypassv2_api('setSwitch', data)
        r = Helpers.process_dev_response(logger, 'toggle_switch', self, r_dict)
        if r is None:
            return False
        self.state.device_status = DeviceStatus.from_bool(toggle)
        self.state.connection_status = ConnectionStatus.ONLINE
        return True

    async def _set_oscillation_state(
        self,
        *,
        vertical: bool | None = None,
        horizontal: bool | None = None,
        left: int | None = None,
        right: int | None = None,
        top: int | None = None,
        bottom: int | None = None,
    ) -> bool:
        """Internal method to set oscillation state.

        Can only set horizontal or vertical at one time. If the vertical argument
        is set (True or False), the horizontal argument and ranges are ignored.
        If the horizontal argument is set (True or False), the vertical
        argument and ranges are ignored. Either vertical or horizontal must be set.

        Args:
            vertical (bool | None): True if setting vertical oscillation.
            horizontal (bool | None): True if setting horizontal oscillation.
            left (int | None): Left range for oscillation (if applicable).
            right (int | None): Right range for oscillation (if applicable).
            top (int | None): Top range for oscillation (if applicable).
            bottom (int | None): Bottom range for oscillation (if applicable).

        Returns:
            bool: true if success.
        """
        if vertical is not None:
            data = {'verticalOscillationState': int(vertical), 'actType': 'default'}
            if vertical is True and (top is not None and bottom is not None):
                data |= {
                    'top': top,
                    'bottom': bottom,
                }
        elif horizontal is not None:
            data = {'horizontalOscillationState': int(horizontal), 'actType': 'default'}
            if horizontal is True and (left is not None and right is not None):
                data |= {
                    'left': left,
                    'right': right,
                }
        else:
            logger.warning('Either vertical or horizontal must be set.')
            return False
        r_dict = await self.call_bypassv2_api('setOscillationStatus', data)
        r = Helpers.process_dev_response(logger, 'set_oscillation_state', self, r_dict)
        if r is None:
            return False
        if vertical is not None:
            self.state.vertical_oscillation_status = DeviceStatus.from_bool(vertical)
        elif horizontal is not None:
            self.state.horizontal_oscillation_status = DeviceStatus.from_bool(horizontal)
        self.state.connection_status = ConnectionStatus.ONLINE
        return True

    async def toggle_vertical_oscillation(self, toggle: bool) -> bool:
        """Toggle vertical oscillation on/off.

        Args:
            toggle (bool): True to enable, False to disable.

        Returns:
            bool: true if success.
        """
        return await self._set_oscillation_state(vertical=toggle)

    async def toggle_horizontal_oscillation(self, toggle: bool) -> bool:
        """Toggle horizontal oscillation on/off.

        Args:
            toggle (bool): True to enable, False to disable.

        Returns:
            bool: true if success.
        """
        return await self._set_oscillation_state(horizontal=toggle)

    async def set_vertical_oscillation_range(
        self, *, top: int = 0, bottom: int = 0
    ) -> bool:
        """Set vertical oscillation range.

        Args:
            top (int): Top range for oscillation.
            bottom (int): Bottom range for oscillation.

        Returns:
            bool: true if success.
        """
        return await self._set_oscillation_state(vertical=True, top=top, bottom=bottom)

    async def set_horizontal_oscillation_range(
        self, *, left: int = 0, right: int = 0
    ) -> bool:
        """Set horizontal oscillation range.

        Args:
            left (int): Left range for oscillation.
            right (int): Right range for oscillation.

        Returns:
            bool: true if success.
        """
        return await self._set_oscillation_state(horizontal=True, left=left, right=right)



"""VeSync API for controlling air purifiers."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from typing_extensions import deprecated

from pyvesync.base_devices.purifier_base import VeSyncPurifier
from pyvesync.const import (
    ConnectionStatus,
    DeviceStatus,
    NightlightModes,
    PurifierAutoPreference,
    PurifierModes,
)
from pyvesync.models.bypass_models import (
    ResultV2GetTimer,
    ResultV2SetTimer,
)
from pyvesync.models.purifier_models import (
    InnerPurifierBaseResult,
    Purifier131Result,
    PurifierCoreDetailsResult,
    PurifierRH131Result,
    PurifierSproutResult,
    PurifierV2EventTiming,
    PurifierV2TimerActionItems,
    PurifierV2TimerPayloadData,
    PurifierVitalDetailsResult,
    RequestPurifier131,
    RequestPurifier131Level,
    RequestPurifier131Mode,
)
from pyvesync.utils.device_mixins import (
    BypassV1Mixin,
    BypassV2Mixin,
    process_bypassv2_result,
)
from pyvesync.utils.helpers import Helpers, Timer

if TYPE_CHECKING:
    from pyvesync import VeSync
    from pyvesync.device_map import PurifierMap
    from pyvesync.models.vesync_models import ResponseDeviceDetailsModel


_LOGGER = logging.getLogger(__name__)


class VeSyncAirBypass(BypassV2Mixin, VeSyncPurifier):
    """Initialize air purifier devices.

    Instantiated by VeSync manager object. Inherits from
    VeSyncBaseDevice class. This is the primary class for most
    air purifiers, using the Bypass V2 API, except the original LV-PUR131S.

    Parameters:
        details (dict): Dictionary of device details
        manager (VeSync): Instantiated VeSync object used to make API calls
        feature_map (PurifierMap): Device map template

    Attributes:
        state (PurifierState): State of the device.
        last_response (ResponseInfo): Last response from API call.
        manager (VeSync): Manager object for API calls.
        device_name (str): Name of device.
        device_image (str): URL for device image.
        cid (str): Device ID.
        connection_type (str): Connection type of device.
        device_type (str): Type of device.
        type (str): Type of device.
        uuid (str): UUID of device, not always present.
        config_module (str): Configuration module of device.
        mac_id (str): MAC ID of device.
        current_firm_version (str): Current firmware version of device.
        device_region (str): Region of device. (US, EU, etc.)
        pid (str): Product ID of device, pulled by some devices on update.
        sub_device_no (int): Sub-device number of device.
        product_type (str): Product type of device.
        features (dict): Features of device.
        modes (list[str]): List of modes supported by the device.
        fan_levels (list[int]): List of fan levels supported by the device.
        nightlight_modes (list[str]): List of nightlight modes supported by the device.
        auto_preferences (list[str]): List of auto preferences supported by the device.

    Notes:
        The `details` attribute holds device information that is updated when
        the `update()` method is called. An example of the `details` attribute:
    ```python
        json.dumps(self.details, indent=4)
        {
            'filter_life': 0,
            'mode': 'manual',
            'level': 0,
            'display': False,
            'child_lock': False,
            'night_light': 'off',
            'air_quality': 0 # air quality level
            'air_quality_value': 0, # PM2.5 value from device,
            'display_forever': False
        }
    ```
    """

    __slots__ = ()

    def __init__(
        self,
        details: ResponseDeviceDetailsModel,
        manager: VeSync,
        feature_map: PurifierMap,
    ) -> None:
        """Initialize VeSync Air Purifier Bypass Base Class."""
        super().__init__(details, manager, feature_map)

    def _set_purifier_state(self, result: PurifierCoreDetailsResult) -> None:
        """Populate PurifierState with details from API response.

        Populates `self.state` and instance variables with device details.

        Args:
            result (InnerPurifierResult): Data model for inner result in purifier
                details response.
        """
        self.state.connection_status = ConnectionStatus.ONLINE
        self.state.device_status = DeviceStatus.ON if result.enabled else DeviceStatus.OFF
        self.state.filter_life = result.filter_life or 0
        self.state.mode = result.mode
        self.state.fan_level = result.level or 0
        self.state.fan_set_level = result.levelNew or 0
        self.state.display_status = (
            DeviceStatus.ON if result.display else DeviceStatus.OFF
        )
        self.state.child_lock = result.child_lock or False
        config = result.configuration
        if config is not None:
            self.state.display_set_status = (
                DeviceStatus.ON if config.display else DeviceStatus.OFF
            )
            self.state.display_forever = config.display_forever
            if config.auto_preference is not None:
                self.state.auto_preference_type = config.auto_preference.type
                self.state.auto_room_size = config.auto_preference.room_size
        if self.supports_air_quality is True:
            self.state.pm25 = result.air_quality_value
            self.state.set_air_quality_level(result.air_quality)
        if result.night_light is not None:
            self.state.nightlight_status = DeviceStatus(result.night_light)

    async def get_details(self) -> None:
        r_dict = await self.call_bypassv2_api('getPurifierStatus')
        resp_model = process_bypassv2_result(
            self, _LOGGER, 'get_details', r_dict, PurifierCoreDetailsResult
        )
        if resp_model is None:
            return
        self._set_purifier_state(resp_model)

    async def get_timer(self) -> Timer | None:
        r_bytes = await self.call_bypassv2_api('getTimer')
        resp_model = process_bypassv2_result(
            self, _LOGGER, 'get_timer', r_bytes, ResultV2GetTimer
        )
        if resp_model is None:
            return None
        timers = resp_model.timers
        if not timers:
            _LOGGER.debug('No timers found')
            self.state.timer = None
            return None
        timer = timers[0]
        self.state.timer = Timer(
            timer_duration=timer.total,
            action=timer.action,
            id=timer.id,
            remaining=timer.remain,
        )
        self.state.device_status = DeviceStatus.ON
        self.state.connection_status = ConnectionStatus.ONLINE
        _LOGGER.debug('Timer found: %s', str(self.state.timer))
        return self.state.timer

    async def set_timer(self, duration: int, action: str | None = None) -> bool:
        action = DeviceStatus.OFF  # No other actions available for this device
        if self.state.device_status != DeviceStatus.ON:
            _LOGGER.debug("Can't set timer when device is off")
        payload_data = {'action': str(action), 'total': duration}
        r_dict = await self.call_bypassv2_api('addTimer', payload_data)
        resp_model = process_bypassv2_result(
            self, _LOGGER, 'set_timer', r_dict, ResultV2SetTimer
        )
        if resp_model is None:
            return False
        self.state.timer = Timer(timer_duration=duration, action='off', id=resp_model.id)
        self.state.device_status = DeviceStatus.ON
        self.state.connection_status = ConnectionStatus.ONLINE
        return True

    async def clear_timer(self) -> bool:
        if self.state.timer is None:
            _LOGGER.debug('No timer to clear, run `get_timer()` to get active timer')
            return False
        payload_data = {'id': self.state.timer.id}

        r_dict = await self.call_bypassv2_api('delTimer', payload_data)
        r = Helpers.process_dev_response(_LOGGER, 'clear_timer', self, r_dict)
        if r is None:
            return False

        _LOGGER.debug('Timer cleared')
        self.state.timer = None
        self.state.device_status = DeviceStatus.ON
        self.state.connection_status = ConnectionStatus.ONLINE
        return True

    async def set_auto_preference(self, preference: str, room_size: int = 800) -> bool:
        if self.state.device_status != DeviceStatus.ON:
            _LOGGER.debug("Can't set auto preference when device is off")
        payload_data = {'type': preference, 'room_size': room_size}
        r_dict = await self.call_bypassv2_api('setAutoPreference', payload_data)
        r = Helpers.process_dev_response(_LOGGER, 'set_auto_preference', self, r_dict)
        if r is None:
            return False
        self.state.connection_status = ConnectionStatus.ONLINE
        self.state.auto_preference_type = preference
        self.state.auto_room_size = room_size
        return True

    async def set_fan_speed(self, speed: None | int = None) -> bool:
        speeds: list = self.fan_levels
        current_speed = self.state.fan_level

        if speed is not None:
            if speed not in speeds:
                _LOGGER.warning(
                    '%s is invalid speed - valid speeds are %s', speed, str(speeds)
                )
                return False
            new_speed = speed
        else:
            new_speed = Helpers.bump_level(current_speed, self.fan_levels)

        data = {
            'id': 0,
            'level': new_speed,
            'type': 'wind',
        }
        r_dict = await self.call_bypassv2_api('setLevel', data)
        r = Helpers.process_dev_response(_LOGGER, 'set_fan_speed', self, r_dict)
        if r is None:
            return False

        self.state.fan_level = new_speed
        self.state.fan_set_level = new_speed
        self.state.mode = PurifierModes.MANUAL  # Set mode to manual to set fan speed
        self.state.device_status = DeviceStatus.ON
        self.state.connection_status = ConnectionStatus.ONLINE
        return True

    async def toggle_child_lock(self, toggle: bool | None = None) -> bool:
        """Toggle child lock.

        Set child lock to on or off. Internal method used by `turn_on_child_lock` and
        `turn_off_child_lock`.

        Args:
            toggle (bool): True to turn child lock on, False to turn off

        Returns:
            bool : True if child lock is set, False if not
        """
        if toggle is None:
            toggle = self.state.child_lock is False
        data = {'child_lock': toggle}

        r_dict = await self.call_bypassv2_api('setChildLock', data)
        r = Helpers.process_dev_response(_LOGGER, 'toggle_child_lock', self, r_dict)
        if r is None:
            return False

        self.state.child_lock = toggle
        self.state.connection_status = ConnectionStatus.ONLINE
        return True

    async def reset_filter(self) -> bool:
        """Reset filter to 100%.

        Returns:
            bool : True if filter is reset, False if not
        """
        r_dict = await self.call_bypassv2_api('resetFilter')
        r = Helpers.process_dev_response(_LOGGER, 'reset_filter', self, r_dict)
        return bool(r)

    @deprecated('Use set_mode(mode: str) instead.')
    async def mode_toggle(self, mode: str) -> bool:
        """Deprecated - Set purifier mode."""
        return await self.set_mode(mode)

    async def set_mode(self, mode: str) -> bool:
        if mode.lower() not in self.modes:
            _LOGGER.warning('Invalid purifier mode used - %s', mode)
            return False

        if mode.lower() == PurifierModes.MANUAL:
            return await self.set_fan_speed(self.state.fan_level or 1)

        data = {
            'mode': mode,
        }
        r_dict = await self.call_bypassv2_api('setPurifierMode', data)

        r = Helpers.process_dev_response(_LOGGER, 'set_mode', self, r_dict)
        if r is None:
            return False

        self.state.mode = mode
        self.state.device_status = DeviceStatus.ON
        self.state.connection_status = ConnectionStatus.ONLINE
        return True

    async def toggle_switch(self, toggle: bool | None = None) -> bool:
        if toggle is None:
            toggle = self.state.device_status != DeviceStatus.ON

        data = {'enabled': toggle, 'id': 0}
        r_dict = await self.call_bypassv2_api('setSwitch', data)
        r = Helpers.process_dev_response(_LOGGER, 'toggle_switch', self, r_dict)
        if r is None:
            return False

        self.state.device_status = DeviceStatus.ON if toggle else DeviceStatus.OFF
        self.state.connection_status = ConnectionStatus.ONLINE
        return True

    async def toggle_display(self, mode: bool) -> bool:
        data = {'state': mode}
        r_dict = await self.call_bypassv2_api('setDisplay', data)
        r = Helpers.process_dev_response(_LOGGER, 'set_display', self, r_dict)
        if r is None:
            return False
        self.state.connection_status = ConnectionStatus.ONLINE
        self.state.display_set_status = DeviceStatus.from_bool(mode)
        return True

    async def set_nightlight_mode(self, mode: str) -> bool:
        if not self.supports_nightlight:
            _LOGGER.debug('Device does not support night light')
            return False
        if mode.lower() not in self.nightlight_modes:
            _LOGGER.warning('Invalid nightlight mode used (on, off or dim)- %s', mode)
            return False

        r_dict = await self.call_bypassv2_api(
            'setNightLight', {'night_light': mode.lower()}
        )
        r = Helpers.process_dev_response(_LOGGER, 'set_night_light', self, r_dict)
        if r is None:
            return False
        self.state.connection_status = ConnectionStatus.ONLINE
        self.state.nightlight_status = NightlightModes(mode.lower())
        return True

    @property
    @deprecated('Use self.state.air_quality instead.')
    def air_quality(self) -> int | None:
        """Get air quality value PM2.5 (ug/m3)."""
        return self.state.pm25

    @property
    @deprecated('Use self.state.fan_level instead.')
    def fan_level(self) -> int | None:
        """Get current fan level."""
        return self.state.fan_level

    @property
    @deprecated('Use self.state.filter_life instead.')
    def filter_life(self) -> int | None:
        """Get percentage of filter life remaining."""
        return self.state.filter_life

    @property
    @deprecated('Use self.state.display_status instead.')
    def display_state(self) -> bool:
        """Get display state.

        See [self.state.display_status][`pyvesync.VeSyncAirBypass.state.display_status`]
        """
        return self.state.display_status == DeviceStatus.ON

    @property
    @deprecated('Use self.state.display_status instead.')
    def screen_status(self) -> bool:
        """Get display status.

        Returns:
            bool : True if display is on, False if off
        """
        return self.state.display_status == DeviceStatus.ON


class VeSyncAirBaseV2(VeSyncAirBypass):
    """Levoit V2 Air Purifier Class.

    Handles the Vital 100S/200S and Sprout Air Purifiers. The
    Sprout purifier has a separate class
    [VeSyncAirSprout][pyvesync.devices.vesyncpurifier.VeSyncAirSprout]
    that overrides the `_set_state` method. Inherits from VeSyncAirBypass
    and VeSyncBaseDevice class. For newer devices that use camel-case API calls.

    Args:
        details (dict): Dictionary of device details
        manager (VeSync): Instantiated VeSync object
        feature_map (PurifierMap): Device map template

    Attributes:
        state (PurifierState): State of the device.
        last_response (ResponseInfo): Last response from API call.
        manager (VeSync): Manager object for API calls.
        device_name (str): Name of device.
        device_image (str): URL for device image.
        cid (str): Device ID.
        connection_type (str): Connection type of device.
        device_type (str): Type of device.
        type (str): Type of device.
        uuid (str): UUID of device, not always present.
        config_module (str): Configuration module of device.
        mac_id (str): MAC ID of device.
        current_firm_version (str): Current firmware version of device.
        device_region (str): Region of device. (US, EU, etc.)
        pid (str): Product ID of device, pulled by some devices on update.
        sub_device_no (int): Sub-device number of device.
        product_type (str): Product type of device.
        features (dict): Features of device.
        modes (list[str]): List of modes supported by the device.
        fan_levels (list[int]): List of fan levels supported by the device.
        nightlight_modes (list[str]): List of nightlight modes supported by the device.
        auto_preferences (list[str]): List of auto preferences supported by the device.
    """

    __slots__ = ()

    def __init__(
        self,
        details: ResponseDeviceDetailsModel,
        manager: VeSync,
        feature_map: PurifierMap,
    ) -> None:
        """Initialize the VeSync Base API V2 Air Purifier Class."""
        super().__init__(details, manager, feature_map)

    def _set_state(self, details: InnerPurifierBaseResult) -> None:
        """Set Purifier state from details response."""
        if not isinstance(details, PurifierVitalDetailsResult):
            _LOGGER.warning('Invalid details model passed to _set_state')
            return
        self.state.connection_status = ConnectionStatus.ONLINE
        self.state.device_status = DeviceStatus.from_int(details.powerSwitch)
        self.state.mode = details.workMode
        self.state.filter_life = details.filterLifePercent
        if details.fanSpeedLevel == 255:  # noqa: PLR2004
            self.state.fan_level = 0
        else:
            self.state.fan_level = details.fanSpeedLevel
        self.state.fan_set_level = details.manualSpeedLevel
        self.state.child_lock = bool(details.childLockSwitch)
        self.state.air_quality_level = details.AQLevel
        self.state.pm25 = details.PM25
        self.state.light_detection_switch = DeviceStatus.from_int(
            details.lightDetectionSwitch
        )
        self.state.light_detection_status = DeviceStatus.from_int(
            details.environmentLightState
        )
        self.state.display_set_status = DeviceStatus.from_int(details.screenSwitch)
        self.state.display_status = DeviceStatus.from_int(details.screenState)
        auto_pref = details.autoPreference
        if auto_pref is not None:
            self.state.auto_preference_type = auto_pref.autoPreferenceType
            self.state.auto_room_size = auto_pref.roomSize

        self.state.pm1 = details.PM1
        self.state.pm10 = details.PM10
        self.state.aq_percent = details.AQPercent
        self.state.fan_rotate_angle = details.fanRotateAngle
        if details.filterOpenState is not None:
            self.state.filter_open_state = bool(details.filterOpenState)
        if details.timerRemain > 0:
            self.state.timer = Timer(details.timerRemain, 'off')

    @property
    @deprecated('Use self.state.fan_set_level instead.')
    def set_speed_level(self) -> int | None:
        """Get set speed level."""
        return self.state.fan_set_level

    @property
    @deprecated("Use self.state.light_detection_switch, this returns 'on' or 'off")
    def light_detection(self) -> bool:
        """Return true if light detection feature is enabled."""
        return self.state.light_detection_switch == DeviceStatus.ON

    @property
    @deprecated("Use self.state.light_detection_status, this returns 'on' or 'off'")
    def light_detection_state(self) -> bool:
        """Return true if light is detected."""
        return self.state.light_detection_status == DeviceStatus.ON

    async def get_details(self) -> None:
        """Build API V2 Purifier details dictionary."""
        r_dict = await self.call_bypassv2_api('getPurifierStatus')
        r_model = process_bypassv2_result(
            self, _LOGGER, 'get_details', r_dict, PurifierVitalDetailsResult
        )
        if r_model is None:
            return

        self._set_state(r_model)

    @deprecated('Use toggle_light_detection(toggle) instead.')
    async def set_light_detection(self, toggle: bool) -> bool:
        """Set light detection feature."""
        return await self.toggle_light_detection(toggle)

    async def toggle_light_detection(self, toggle: bool | None = None) -> bool:
        """Enable/Disable Light Detection Feature."""
        if bool(self.state.light_detection_status) == toggle:
            _LOGGER.debug(
                'Light detection is already %s', self.state.light_detection_status
            )
            return True

        if toggle is None:
            toggle = not bool(self.state.light_detection_status)
        payload_data = {'lightDetectionSwitch': int(toggle)}
        r_dict = await self.call_bypassv2_api('setLightDetection', payload_data)
        r = Helpers.process_dev_response(_LOGGER, 'set_light_detection', self, r_dict)
        if r is None:
            return False

        self.state.light_detection_switch = DeviceStatus.from_bool(toggle)
        self.state.connection_status = ConnectionStatus.ONLINE
        return True

    async def toggle_switch(self, toggle: bool | None = None) -> bool:
        if toggle is None:
            toggle = not bool(self.state.device_status)
        if toggle == bool(self.state.device_status):
            _LOGGER.debug('Purifier is already %s', self.state.device_status)
            return True

        payload_data = {'powerSwitch': int(toggle), 'switchIdx': 0}
        r_dict = await self.call_bypassv2_api('setSwitch', payload_data)
        r = Helpers.process_dev_response(_LOGGER, 'toggle_switch', self, r_dict)
        if r is None:
            return False

        self.state.device_status = DeviceStatus.from_bool(toggle)
        self.state.connection_status = ConnectionStatus.ONLINE
        return True

    async def toggle_child_lock(self, toggle: bool | None = None) -> bool:
        """Levoit 100S/200S set Child Lock.

        Parameters:
            toggle (bool): True to turn child lock on, False to turn off

        Returns:
            bool : True if successful, False if not
        """
        if toggle is None:
            toggle = not bool(self.state.child_lock)
        payload_data = {'childLockSwitch': int(toggle)}
        r_dict = await self.call_bypassv2_api('setChildLock', payload_data)

        r = Helpers.process_dev_response(_LOGGER, 'toggle_child_lock', self, r_dict)
        if r is None:
            return False

        self.state.child_lock = toggle
        self.state.connection_status = ConnectionStatus.ONLINE
        return True

    async def toggle_display(self, mode: bool) -> bool:
        if bool(self.state.light_detection_status):
            _LOGGER.error('Cannot set display when light detection is enabled')
            return False

        if bool(self.state.display_set_status) == mode:
            _LOGGER.debug('Display is already %s', mode)
            return True

        payload_data = {'screenSwitch': int(mode)}
        r_dict = await self.call_bypassv2_api('setDisplay', payload_data)

        r = Helpers.process_dev_response(_LOGGER, 'set_display', self, r_dict)
        if r is None:
            return False

        self.state.display_set_status = DeviceStatus.from_bool(mode)
        self.state.connection_status = ConnectionStatus.ONLINE
        return True

    async def set_timer(self, duration: int, action: str | None = None) -> bool:
        action = DeviceStatus.OFF  # No other actions available for this device
        if action not in [DeviceStatus.ON, DeviceStatus.OFF]:
            _LOGGER.warning('Invalid action for timer')
            return False

        method = 'powerSwitch'
        action_int = 1 if action == DeviceStatus.ON else 0
        action_item = PurifierV2TimerActionItems(type=method, act=action_int)
        timing = PurifierV2EventTiming(clkSec=duration)
        payload_data = PurifierV2TimerPayloadData(
            enabled=True,
            startAct=[action_item],
            tmgEvt=timing,
        )

        r_dict = await self.call_bypassv2_api('addTimerV2', payload_data.to_dict())
        r = Helpers.process_dev_response(_LOGGER, 'set_timer', self, r_dict)
        if r is None:
            return False

        r_model = ResultV2SetTimer.from_dict(r)

        self.state.timer = Timer(duration, action=action, id=r_model.id)
        self.state.connection_status = ConnectionStatus.ONLINE
        return True

    async def clear_timer(self) -> bool:
        if self.state.timer is None:
            _LOGGER.warning('No timer found, run get_timer() to retrieve timer.')
            return False

        payload_data = {'id': self.state.timer.id, 'subDeviceNo': 0}
        r_dict = await self.call_bypassv2_api('delTimerV2', payload_data)
        r = Helpers.process_dev_response(_LOGGER, 'clear_timer', self, r_dict)
        if r is None:
            return False

        self.state.timer = None
        self.state.connection_status = ConnectionStatus.ONLINE
        return True

    async def set_auto_preference(
        self, preference: str = PurifierAutoPreference.DEFAULT, room_size: int = 600
    ) -> bool:
        """Set Levoit Vital 100S/200S auto mode.

        Parameters:
            preference (str | None | PurifierAutoPreference):
                Preference for auto mode, default 'default' (default, efficient, quiet)
            room_size (int | None):
                Room size in square feet, by default 600
        """
        if preference not in self.auto_preferences:
            _LOGGER.warning(
                '%s is invalid preference -'
                ' valid preferences are default, efficient, quiet',
                preference,
            )
            return False
        payload_data = {'autoPreference': preference, 'roomSize': room_size}
        r_dict = await self.call_bypassv2_api('setAutoPreference', payload_data)
        r = Helpers.process_dev_response(_LOGGER, 'set_auto_preference', self, r_dict)
        if r is None:
            return False

        self.state.auto_preference_type = preference
        self.state.auto_room_size = room_size
        self.state.connection_status = ConnectionStatus.ONLINE
        return True

    async def set_fan_speed(self, speed: None | int = None) -> bool:
        if speed is not None:
            if speed not in self.fan_levels:
                _LOGGER.warning(
                    '%s is invalid speed - valid speeds are %s',
                    speed,
                    str(self.fan_levels),
                )
                return False
            new_speed = speed
        elif self.state.fan_level is None:
            new_speed = self.fan_levels[0]
        else:
            new_speed = Helpers.bump_level(self.state.fan_level, self.fan_levels)

        payload_data = {'levelIdx': 0, 'manualSpeedLevel': new_speed, 'levelType': 'wind'}
        r_dict = await self.call_bypassv2_api('setLevel', payload_data)
        r = Helpers.process_dev_response(_LOGGER, 'set_fan_speed', self, r_dict)
        if r is None:
            return False

        self.state.fan_set_level = new_speed
        self.state.fan_level = new_speed
        self.state.mode = PurifierModes.MANUAL
        self.state.device_status = DeviceStatus.ON
        self.state.connection_status = ConnectionStatus.ONLINE
        return True

    async def set_mode(self, mode: str) -> bool:
        if mode.lower() not in self.modes:
            _LOGGER.warning('Invalid purifier mode used - %s', mode)
            return False

        # Call change_fan_speed if mode is set to manual
        if mode == PurifierModes.MANUAL:
            if self.state.fan_set_level is None or self.state.fan_level == 0:
                return await self.set_fan_speed(1)
            return await self.set_fan_speed(self.state.fan_set_level)

        payload_data = {'workMode': mode}
        r_dict = await self.call_bypassv2_api('setPurifierMode', payload_data)
        r = Helpers.process_dev_response(_LOGGER, 'mode_toggle', self, r_dict)
        if r is None:
            return False

        self.state.mode = mode
        self.state.connection_status = ConnectionStatus.ONLINE
        self.state.device_status = DeviceStatus.ON
        return True


class VeSyncAirSprout(VeSyncAirBaseV2):  # pylint: disable=too-many-ancestors
    """Class for the Sprout Air Purifier.

    Inherits from VeSyncAirBaseV2 class and overrides
    the _set_state method. See the
    [VeSyncAirBaseV2][pyvesync.devices.vesyncpurifier.VeSyncAirBaseV2]
    class for more information.

    Args:
        details (dict): Dictionary of device details
        manager (VeSync): Instantiated VeSync object
        feature_map (PurifierMap): Device map template

    Attributes:
        state (PurifierState): State of the device.
        last_response (ResponseInfo): Last response from API call.
        manager (VeSync): Manager object for API calls.
        device_name (str): Name of device.
        device_image (str): URL for device image.
        cid (str): Device ID.
        connection_type (str): Connection type of device.
        device_type (str): Type of device.
        type (str): Type of device.
        uuid (str): UUID of device, not always present.
        config_module (str): Configuration module of device.
        mac_id (str): MAC ID of device.
        current_firm_version (str): Current firmware version of device.
        device_region (str): Region of device. (US, EU, etc.)
        pid (str): Product ID of device, pulled by some devices on update.
        sub_device_no (int): Sub-device number of device.
        product_type (str): Product type of device.
        features (dict): Features of device.
        modes (list[str]): List of modes supported by the device.
        fan_levels (list[int]): List of fan levels supported by the device.
        nightlight_modes (list[str]): List of nightlight modes supported by the device.
        auto_preferences (list[str]): List of auto preferences supported by the device.
    """

    def __init__(
        self,
        details: ResponseDeviceDetailsModel,
        manager: VeSync,
        feature_map: PurifierMap,
    ) -> None:
        """Initialize air purifier class."""
        super().__init__(details, manager, feature_map)

    def _set_state(self, details: InnerPurifierBaseResult) -> None:
        """Set Purifier state from details response."""
        if not isinstance(details, PurifierSproutResult):
            _LOGGER.warning('Invalid details model passed to _set_state')
            return
        self.state.connection_status = ConnectionStatus.ONLINE
        self.state.device_status = DeviceStatus.from_int(details.powerSwitch)
        self.state.mode = details.workMode
        if details.fanSpeedLevel == 255:  # noqa: PLR2004
            self.state.fan_level = 0
        else:
            self.state.fan_level = details.fanSpeedLevel
        self.state.fan_set_level = details.manualSpeedLevel
        self.state.child_lock = bool(details.childLockSwitch)
        self.state.air_quality_level = details.AQLevel
        self.state.pm25 = details.PM25
        self.state.pm1 = details.PM1
        self.state.pm10 = details.PM10
        self.state.aq_percent = details.AQI
        self.state.display_set_status = DeviceStatus.from_int(details.screenSwitch)
        self.state.display_status = DeviceStatus.from_int(details.screenState)
        auto_pref = details.autoPreference
        if auto_pref is not None:
            self.state.auto_preference_type = auto_pref.autoPreferenceType
            self.state.auto_room_size = auto_pref.roomSize
        self.state.humidity = details.humidity
        self.state.temperature = int((details.temperature or 0) / 10)
        self.state.pm1 = details.PM1
        self.state.pm10 = details.PM10
        self.state.pm25 = details.PM25
        self.state.voc = details.VOC
        self.state.co2 = details.CO2
        if details.nightlight is not None:
            self.state.nightlight_status = DeviceStatus.from_int(
                details.nightlight.nightLightSwitch
            )
            self.state.nightlight_brightness = details.nightlight.brightness

    async def get_details(self) -> None:
        """Build API V2 Purifier details dictionary."""
        r_dict = await self.call_bypassv2_api('getPurifierStatus')
        r_model = process_bypassv2_result(
            self, _LOGGER, 'get_details', r_dict, PurifierSproutResult
        )
        if r_model is None:
            return

        self._set_state(r_model)


class VeSyncAir131(BypassV1Mixin, VeSyncPurifier):
    """Levoit Air Purifier Class.

    Class for LV-PUR131S, using BypassV1 API.

    Attributes:
        state (PurifierState): State of the device.
        last_response (ResponseInfo): Last response from API call.
        manager (VeSync): Manager object for API calls.
        device_name (str): Name of device.
        device_image (str): URL for device image.
        cid (str): Device ID.
        connection_type (str): Connection type of device.
        device_type (str): Type of device.
        type (str): Type of device.
        uuid (str): UUID of device, not always present.
        config_module (str): Configuration module of device.
        mac_id (str): MAC ID of device.
        current_firm_version (str): Current firmware version of device.
        device_region (str): Region of device. (US, EU, etc.)
        pid (str): Product ID of device, pulled by some devices on update.
        sub_device_no (int): Sub-device number of device.
        product_type (str): Product type of device.
        features (dict): Features of device.
        modes (list[str]): List of modes supported by the device.
        fan_levels (list[int]): List of fan levels supported by the device.
        nightlight_modes (list[str]): List of nightlight modes supported by the device.
        auto_preferences (list[str]): List of auto preferences supported by the device.
    """

    __slots__ = ()

    def __init__(
        self,
        details: ResponseDeviceDetailsModel,
        manager: VeSync,
        feature_map: PurifierMap,
    ) -> None:
        """Initialize air purifier class."""
        super().__init__(details, manager, feature_map)

    def _set_state(self, details: Purifier131Result) -> None:
        """Set state from purifier API get_details() response."""
        self.state.device_status = details.deviceStatus
        self.state.connection_status = details.connectionStatus
        self.state.active_time = details.activeTime
        if details.filterLife is not None:
            self.state.filter_life = details.filterLife.percent
        self.state.display_status = DeviceStatus(details.screenStatus)
        self.state.display_set_status = details.screenStatus
        self.state.mode = details.mode
        self.state.fan_level = details.level or 0
        self.state.fan_set_level = details.level or 0
        self.state.set_air_quality_level(details.airQuality)

    async def get_details(self) -> None:
        r_dict = await self.call_bypassv1_api(
            RequestPurifier131, method='deviceDetail', endpoint='deviceDetail'
        )
        r = Helpers.process_dev_response(_LOGGER, 'get_details', self, r_dict)
        if r is None:
            return

        r_model = Purifier131Result.from_dict(r.get('result', {}))
        self._set_state(r_model)

    async def toggle_display(self, mode: bool) -> bool:
        update_dict = {'status': 'on' if mode else 'off'}
        r_dict = await self.call_bypassv1_api(
            RequestPurifier131,
            method='airPurifierScreenCtl',
            endpoint='airPurifierScreenCtl',
            update_dict=update_dict,
        )
        r = Helpers.process_dev_response(_LOGGER, 'toggle_display', self, r_dict)
        if r is None:
            return False

        self.state.display_set_status = DeviceStatus.from_bool(mode)
        self.state.display_status = DeviceStatus.from_bool(mode)
        self.state.connection_status = ConnectionStatus.ONLINE
        return True

    async def toggle_switch(self, toggle: bool | None = None) -> bool:
        if toggle is None:
            toggle = self.state.device_status != DeviceStatus.ON

        update_dict = {'status': DeviceStatus.from_bool(toggle).value}
        r_dict = await self.call_bypassv1_api(
            RequestPurifier131,
            method='airPurifierPowerSwitchCtl',
            endpoint='airPurifierPowerSwitchCtl',
            update_dict=update_dict,
        )
        r = Helpers.process_dev_response(_LOGGER, 'toggle_switch', self, r_dict)
        if r is None:
            return False

        self.state.device_status = DeviceStatus.from_bool(toggle)
        self.state.connection_status = ConnectionStatus.ONLINE
        return True

    async def set_fan_speed(self, speed: int | None = None) -> bool:
        current_speed = self.state.fan_set_level or 0

        if speed is not None:
            if speed not in self.fan_levels:
                _LOGGER.warning(
                    '%s is invalid speed - valid speeds are %s',
                    speed,
                    str(self.fan_levels),
                )
                return False
            new_speed = speed
        else:
            new_speed = Helpers.bump_level(current_speed, self.fan_levels)

        update_dict = {'level': new_speed}
        r_dict = await self.call_bypassv1_api(
            RequestPurifier131Level,
            method='airPurifierSpeedCtl',
            endpoint='airPurifierSpeedCtl',
            update_dict=update_dict,
        )
        r = Helpers.process_dev_response(_LOGGER, 'set_fan_speed', self, r_dict)
        if r is None:
            return False

        self.state.fan_level = new_speed
        self.state.fan_set_level = new_speed
        self.state.connection_status = 'online'
        self.state.mode = PurifierModes.MANUAL
        return True

    async def set_mode(self, mode: str) -> bool:
        if mode not in self.modes:
            _LOGGER.warning('Invalid purifier mode used - %s', mode)
            return False

        if mode == PurifierModes.MANUAL:
            set_level = (
                1 if self.state.fan_set_level in [0, None] else self.state.fan_set_level
            )
            return await self.set_fan_speed(set_level)

        update_dict = {'mode': mode}
        r_dict = await self.call_bypassv1_api(
            RequestPurifier131Mode,
            method='airPurifierRunModeCtl',
            endpoint='airPurifierRunModeCtl',
            update_dict=update_dict,
        )
        r = Helpers.process_dev_response(_LOGGER, 'mode_toggle', self, r_dict)
        if r is None:
            return False

        self.state.mode = mode
        self.state.connection_status = ConnectionStatus.ONLINE
        return True


class VeSyncAirRH131(BypassV2Mixin, VeSyncPurifier):
    """Levoit Air Purifier Class.

    Class for LV-PUR131S, using BypassV1 API.

    Attributes:
        state (PurifierState): State of the device.
        last_response (ResponseInfo): Last response from API call.
        manager (VeSync): Manager object for API calls.
        device_name (str): Name of device.
        device_image (str): URL for device image.
        cid (str): Device ID.
        connection_type (str): Connection type of device.
        device_type (str): Type of device.
        type (str): Type of device.
        uuid (str): UUID of device, not always present.
        config_module (str): Configuration module of device.
        mac_id (str): MAC ID of device.
        current_firm_version (str): Current firmware version of device.
        device_region (str): Region of device. (US, EU, etc.)
        pid (str): Product ID of device, pulled by some devices on update.
        sub_device_no (int): Sub-device number of device.
        product_type (str): Product type of device.
        features (dict): Features of device.
        modes (list[str]): List of modes supported by the device.
        fan_levels (list[int]): List of fan levels supported by the device.
        nightlight_modes (list[str]): List of nightlight modes supported by the device.
        auto_preferences (list[str]): List of auto preferences supported by the device.
    """

    __slots__ = ()

    def __init__(
        self,
        details: ResponseDeviceDetailsModel,
        manager: VeSync,
        feature_map: PurifierMap,
    ) -> None:
        """Initialize air purifier class."""
        super().__init__(details, manager, feature_map)

    def _set_state(self, details: PurifierRH131Result) -> None:
        """Set state from purifier API get_details() response."""
        self.state.device_status = DeviceStatus.from_bool(details.enabled)
        self.state.connection_status = ConnectionStatus.ONLINE
        if details.filter_life is not None:
            self.state.filter_life = details.filter_life
        self.state.display_status = DeviceStatus.from_bool(details.display)
        self.state.display_set_status = DeviceStatus.from_bool(details.display)
        self.state.mode = details.mode
        self.state.fan_level = details.level or 0
        self.state.fan_set_level = details.level or 0
        self.state.set_air_quality_level(details.air_quality or 0)
        self.state.child_lock = details.child_lock

    async def get_details(self) -> None:
        r_dict = await self.call_bypassv2_api('getPurifierStatus')
        r_model = process_bypassv2_result(
            self, _LOGGER, 'get_details', r_dict, PurifierRH131Result
        )
        if r_model is None:
            return

        self._set_state(r_model)

    async def toggle_switch(self, toggle: bool | None = None) -> bool:
        if toggle is None:
            toggle = self.state.device_status != DeviceStatus.ON

        payload_data = {'enabled': toggle, 'id': 0}
        r_dict = await self.call_bypassv2_api('setSwitch', payload_data)
        r = Helpers.process_dev_response(_LOGGER, 'toggle_switch', self, r_dict)
        if r is None:
            return False

        self.state.device_status = DeviceStatus.from_bool(toggle)
        self.state.connection_status = ConnectionStatus.ONLINE
        return True

    async def toggle_display(self, mode: bool) -> bool:
        update_dict = {'state': mode}
        r_dict = await self.call_bypassv2_api('setDisplay', update_dict)
        r = Helpers.process_dev_response(_LOGGER, 'toggle_display', self, r_dict)
        if r is None:
            return False

        self.state.display_set_status = DeviceStatus.from_bool(mode)
        self.state.display_status = DeviceStatus.from_bool(mode)
        self.state.connection_status = ConnectionStatus.ONLINE
        return True

    async def set_mode(self, mode: str) -> bool:
        if mode not in self.modes:
            _LOGGER.warning('Invalid purifier mode used - %s', mode)
            return False

        payload_data = {'mode': mode}
        r_dict = await self.call_bypassv2_api('setPurifierMode', payload_data)
        r = Helpers.process_dev_response(_LOGGER, 'set_mode', self, r_dict)
        if r is None:
            return False

        self.state.mode = mode
        self.state.connection_status = ConnectionStatus.ONLINE
        return True

    async def set_fan_speed(self, speed: int | None = None) -> bool:
        if speed is not None:
            if speed not in self.fan_levels:
                _LOGGER.warning(
                    '%s is invalid speed - valid speeds are %s',
                    speed,
                    str(self.fan_levels),
                )
                return False
            new_speed = speed
        else:
            new_speed = Helpers.bump_level(self.state.fan_level or 0, self.fan_levels)

        payload_data = {'level': new_speed, 'id': 0, 'type': 'wind'}

        r_dict = await self.call_bypassv2_api('setLevel', payload_data)

        r = Helpers.process_dev_response(_LOGGER, 'set_fan_speed', self, r_dict)
        if r is None:
            return False

        self.state.fan_level = new_speed
        self.state.fan_set_level = new_speed
        self.state.connection_status = ConnectionStatus.ONLINE
        self.state.mode = PurifierModes.MANUAL
        return True

    async def toggle_child_lock(self, toggle: bool | None = None) -> bool:
        if toggle is None:
            toggle = not bool(self.state.child_lock)
        payload_data = {'child_lock': toggle}
        r_dict = await self.call_bypassv2_api('setChildLock', payload_data)

        r = Helpers.process_dev_response(_LOGGER, 'toggle_child_lock', self, r_dict)
        if r is None:
            return False

        self.state.child_lock = toggle
        self.state.connection_status = ConnectionStatus.ONLINE
        return True



"""
Thermostat Device API Responses

THERMOSTATS variable is a list of setup_entry's from the device_map

DETAILS_RESPONSES variable is a dictionary of responses from the API
for get_details() methods.  The keys are the device types and the
values are the responses.  The responses are tuples of (response, status)

METHOD_RESPONSES variable is a defaultdict of responses from the API. This is
the FunctionResponse variable from the utils module in the tests dir.
The default response is a tuple with the value ({"code": 0, "msg": "success"}, 200).

The values of METHOD_RESPONSES can be a function that takes a single argument or
a static value. The value is checked if callable at runtime and if so, it is called
with the provided argument. If not callable, the value is returned as is.

METHOD_RESPONSES = {
    'DEV_TYPE': defaultdict(
        lambda: {"code": 0, "msg": "success"}
    )
}

# For a function to handle the response
def status_response(request_body=None):
    # do work with request_body
    return request_body, 200

METHOD_RESPONSES['DEV_TYPE']['set_status'] = status_response

# To change the default value for a device type

METHOD_RESPONSES['DEVTYPE'].default_factory = lambda: {"code": 0, "msg": "success"}
"""
from copy import deepcopy
from pyvesync.device_map import thermostat_modules
from pyvesync.const import ThermostatConst
from defaults import (
    build_bypass_v2_response,
    FunctionResponsesV2,
)

THERMOSTATS = [m.setup_entry for m in thermostat_modules]
THERMOSTATS_NUM = len(THERMOSTATS)


class ThermostatDefaults:
    current_temp = 65
    target_temp = 75
    humidity = 45
    work_mode = ThermostatConst.WorkMode.HEAT
    work_status = ThermostatConst.WorkStatus.HEATING
    eco_mode = ThermostatConst.EcoType.BALANCE
    temp_unit = "f"
    schedule_enabled = "on"
    away_mode = "off"
    child_lock = "off"
    screen_display = "on"
    outdoor_temp = 600
    outdoor_humidity = 300
    filter_life = 100
    battery_level = 100
    hvac_state = "idle"
    compressor_state = "off"
    fan_state = "off"
    fan_speed = "auto"
    fan_speed_level = 1
    swing_mode = "off"
    aux_heat_state = "off"


THERMOSTAT_DETAILS: dict[str, dict] = {
    "LTM-A401S-WUS": {
        "supportMode": [0, 1, 2, 3, 5],
        "workMode": 3,
        "workStatus": 2,
        "fanMode": 1,
        "fanStatus": 1,
        "routineRunningId": 2,
        "tempUnit": "f",
        "temperature": 75.570000,
        "humidity": 56,
        "heatToTemp": 60,
        "coolToTemp": 70,
        "lockStatus": False,
        "scheduleOrHold": 2,
        "holdEndTime": 1696033274,
        "holdOption": 5,
        "deadband": 4,
        "safetyTempLow": "40.00",
        "safetyTempHigh": "100.00",
        "ecoType": 3,
        "alertStatus": 2,
        "routines": [
            {"name": "Home", "routineId": 2},
            {"name": "Away", "routineId": 1},
            {"name": "Sleep", "routineId": 3},
        ],
    },
}
"""Dictionary with setup_entry as the key and the value being the result object of
the device response from the get_details API for thermostats."""


DEVICE_DETAILS = {
    "LTM-A401S-WUS": build_bypass_v2_response(inner_result=THERMOSTAT_DETAILS["LTM-A401S-WUS"]),
}

METHOD_RESPONSES = {
    "LTM-A401S-WUS": deepcopy(FunctionResponsesV2)
}



"""
Light Bulbs Device API Responses

OUTLET variable is a list of device types

DETAILS_RESPONSES variable is a dictionary of responses from the API
for get_details() methods.  The keys are the device types and the
values are the responses.  The responses are tuples of (response, status)

METHOD_RESPONSES variable is a defaultdict of responses from the API. This is
the FunctionResponse variable from the utils module in the tests dir.
The default response is a tuple with the value ({"code": 0, "msg": "success"}, 200).

The values of METHOD_RESPONSES can be a function that takes a single argument or
a static value. The value is checked if callable at runtime and if so, it is called
with the provided argument. If not callable, the value is returned as is.

METHOD_RESPONSES = {
    'ESL100': defaultdict(
        lambda: ({"code": 0, "msg": "success"}, 200))
    )
}

# For a function to handle the response
def status_response(request_body=None):
    # do work with request_body
    return request_body, 200

METHOD_RESPONSES['ESL100']['set_status'] = status_response

# To change the default value for a device type

METHOD_RESPONSES['XYD0001'].default_factory = lambda: ({"code": 0, "msg": "success"}, 200)

"""
from copy import deepcopy
from pyvesync.device_map import bulb_modules
from pyvesync.const import DeviceStatus, ConnectionStatus
from defaults import (
    TestDefaults,
    FunctionResponsesV2,
    FunctionResponsesV1,
    build_bypass_v1_response,
    build_bypass_v2_response,
)

# BULBS = ['ESL100', 'ESL100CW', 'ESL100MC', 'XYD0001']
BULBS = [m.setup_entry for m in bulb_modules]
BULBS_NUM = len(BULBS)


BULB_DETAILS: dict[str, dict[str, str | float | dict | None]] = {
    "ESL100": {
        "deviceName": "Dimmable",
        "name": "Dimmable",
        "brightNess": str(TestDefaults.brightness),
        "deviceStatus": DeviceStatus.ON.value,
        "activeTime": TestDefaults.active_time,
        "defaultDeviceImg": "",
        "timer": None,
        "scheduleCount": 0,
        "away": None,
        "schedule": None,
        "ownerShip": "1",
        "deviceImg": "",
        "connectionStatus": ConnectionStatus.ONLINE.value,
    },
    "ESL100CW": {
        "light": {
            "action": DeviceStatus.ON.value,
            "brightness": TestDefaults.brightness,
            "colorTempe": TestDefaults.color_temp,
        }
    },
    "ESL100MC": {
        "action": DeviceStatus.ON.value,
        "brightness": TestDefaults.brightness,
        "colorMode": "color",
        "speed": 0,
        "red": TestDefaults.color.rgb.red,
        "green": TestDefaults.color.rgb.green,
        "blue": TestDefaults.color.rgb.blue,
    },
    "XYD0001": {
        "enabled": DeviceStatus.ON.value,
        "colorMode": "color",
        "brightness": TestDefaults.brightness,
        "colorTemp": TestDefaults.color_temp,
        "hue": TestDefaults.color.hsv.hue * 27.7778,
        "saturation": TestDefaults.color.hsv.saturation * 100,
        "value": TestDefaults.color.hsv.value,
    },
}

# class BulbDetails:
#     details_esl100 = ({
#         'code': 0,
#         'msg': None,
#         'deviceStatus': 'on',
#         'connectionStatus': 'online',
#         'name': Defaults.name('ESL100'),
#         'brightNess': Defaults.brightness,
#         'timer': None,
#         'away': None,
#         'schedule': None,
#         'ownerShip': '1',
#         'scheduleCount': 0,
#     }, 200)

#     details_esl100cw = {
# 		"light": {
# 			"action": "off",
# 			"brightness": 4,
# 			"colorTempe": 50
# 		}
# 	}

#     details_esl100mc = ({
#         "traceId": Defaults.trace_id,
#         "code": 0,
#         "msg": "request success",
#         "result": {
#             "traceId": Defaults.trace_id,
#             "code": 0,
#             "result": {
#                 "action": "on",
#                 "brightness": Defaults.brightness,
#                 "colorMode": "color",
#                 "speed": 0,
#                 "red": Defaults.color.rgb.red,
#                 "green": Defaults.color.rgb.green,
#                 "blue": Defaults.color.rgb.blue,
#             }
#         }
#     }, 200)

#     details_valceno = (
#             {
#                 "traceId": TRACE_ID,
#                 "code": 0,
#                 "msg": "request success",
#                 "result": {
#                     "traceId": TRACE_ID,
#                     "code": 0,
#                     "result": {
#                         "enabled": "on",
#                         "colorMode": "color",
#                         "brightness": Defaults.brightness,
#                         "colorTemp": Defaults.color_temp,
#                         "hue": Defaults.color.hsv.hue*27.7778,
#                         "saturation": Defaults.color.hsv.saturation*100,
#                         "value": Defaults.color.hsv.value,
#                     }
#                 }
#             }, 200
#         )


DETAILS_RESPONSES = {
    'ESL100': build_bypass_v1_response(result_dict=BULB_DETAILS['ESL100']),
    'ESL100CW': build_bypass_v1_response(result_dict=BULB_DETAILS['ESL100CW']),
    'ESL100MC': build_bypass_v2_response(inner_result=BULB_DETAILS['ESL100MC']),
    'XYD0001': build_bypass_v2_response(inner_result=BULB_DETAILS['XYD0001']),
}


def valceno_set_status_response(kwargs=None):
    default_resp = {
        "traceId": TestDefaults.trace_id,
        "code": 0,
        "msg": "request success",
        "result": {
            "traceId": TestDefaults.trace_id,
            "code": 0,
            "result": {
                "enabled": "on",
                "colorMode": "hsv",
                "brightness": TestDefaults.brightness,
                "colorTemp": TestDefaults.color_temp,
                "hue": TestDefaults.color.hsv.hue*27.7778,
                "saturation": TestDefaults.color.hsv.saturation*100,
                "value": TestDefaults.color.hsv.value
            }
        }
    }
    if isinstance(kwargs, dict):
        if kwargs.get('hue') is not None:
            default_resp['result']['result']['hue'] = kwargs['hue'] * 27.7778
        if kwargs.get('saturation') is not None:
            default_resp['result']['result']['saturation'] = kwargs['saturation'] * 100
        if kwargs.get('value') is not None:
            default_resp['result']['result']['value'] = kwargs['value']
    return default_resp


METHOD_RESPONSES = {
    'ESL100': deepcopy(FunctionResponsesV1),
    'ESL100CW': deepcopy(FunctionResponsesV1),
    'ESL100MC': deepcopy(FunctionResponsesV2),
    'XYD0001': deepcopy(FunctionResponsesV2),
}

# METHOD_RESPONSES['XYD0001'].default_factory = lambda: ({
#                                                 "traceId": Defaults.trace_id,
#                                                 "code": 0,
#                                                 "msg": "request success",
#                                                 "result": {
#                                                     "traceId": Defaults.trace_id,
#                                                     "code": 0
#                                                 }
#                                             }, 200)

# METHOD_RESPONSES['ESL100MC'].default_factory = lambda: ({
#                                                 "traceId": Defaults.trace_id,
#                                                 "code": 0,
#                                                 "msg": "request success",
#                                                 "result": {
#                                                     "traceId": Defaults.trace_id,
#                                                     "code": 0
#                                                 }
#                                             }, 200)

XYD0001_RESP = {
    'set_brightness': valceno_set_status_response,
    'set_color_temp': valceno_set_status_response,
    'set_hsv': valceno_set_status_response,
    'set_rgb': valceno_set_status_response,
}

METHOD_RESPONSES['XYD0001'].update(XYD0001_RESP)



"""VeSync Kitchen Devices.

The Cosori 3.7 and 5.8 Quart Air Fryer has several methods and properties that can be
used to monitor and control the device.

To maintain consistency of state, the update() method is called after each of the methods
that change the state of the device.

There is also an instance attribute that can be set `VeSyncAirFryer158.refresh_interval`
that will set the interval in seconds that the state of the air fryer should be updated
before a method that changes state is called. This is an additional API call but is
necessary to maintain state, especially when trying to `pause` or `resume` the device.
Defaults to 60 seconds but can be set via:

```python
# Change to 120 seconds before status is updated between calls
VeSyncAirFryer158.refresh_interval = 120

# Set status update before every call
VeSyncAirFryer158.refresh_interval = 0

# Disable status update before every call
VeSyncAirFryer158.refresh_interval = -1
```

"""

from __future__ import annotations

import logging
import time
from typing import TYPE_CHECKING, TypeVar

from typing_extensions import deprecated

from pyvesync.base_devices import FryerState, VeSyncFryer
from pyvesync.const import AIRFRYER_PID_MAP, ConnectionStatus, DeviceStatus
from pyvesync.utils.errors import VeSyncError
from pyvesync.utils.helpers import Helpers
from pyvesync.utils.logs import LibraryLogger

if TYPE_CHECKING:
    from pyvesync import VeSync
    from pyvesync.device_map import AirFryerMap
    from pyvesync.models.vesync_models import ResponseDeviceDetailsModel

T = TypeVar('T')

logger = logging.getLogger(__name__)


# Status refresh interval in seconds
# API calls outside of interval are automatically refreshed
# Set VeSyncAirFryer158.refresh_interval to 0 to refresh every call
# Set to None or -1 to disable auto-refresh
REFRESH_INTERVAL = 60

RECIPE_ID = 1
RECIPE_TYPE = 3
CUSTOM_RECIPE = 'Manual Cook'
COOK_MODE = 'custom'


class AirFryer158138State(FryerState):
    """Dataclass for air fryer status.

    Attributes:
        active_time (int): Active time of device, defaults to None.
        connection_status (str): Connection status of device.
        device (VeSyncBaseDevice): Device object.
        device_status (str): Device status.
        features (dict): Features of device.
        last_update_ts (int): Last update timestamp of device, defaults to None.
        ready_start (bool): Ready start status of device, defaults to False.
        preheat (bool): Preheat status of device, defaults to False.
        cook_status (str): Cooking status of device, defaults to None.
        current_temp (int): Current temperature of device, defaults to None.
        cook_set_temp (int): Cooking set temperature of device, defaults to None.
        last_timestamp (int): Last timestamp of device, defaults to None.
        preheat_set_time (int): Preheat set time of device, defaults to None.
        preheat_last_time (int): Preheat last time of device, defaults to None.
        _temp_unit (str): Temperature unit of device, defaults to None.
    """

    __slots__ = (
        '_temp_unit',
        'cook_last_time',
        'cook_set_temp',
        'cook_set_time',
        'cook_status',
        'current_temp',
        'last_timestamp',
        'max_temp_c',
        'max_temp_f',
        'min_temp_c',
        'min_temp_f',
        'preheat',
        'preheat_last_time',
        'preheat_set_time',
        'ready_start',
    )

    def __init__(
        self,
        device: VeSyncAirFryer158,
        details: ResponseDeviceDetailsModel,
        feature_map: AirFryerMap,
    ) -> None:
        """Init the Air Fryer 158 class."""
        super().__init__(device, details, feature_map)
        self.device: VeSyncFryer = device
        self.features: list[str] = feature_map.features
        self.min_temp_f: int = feature_map.temperature_range_f[0]
        self.max_temp_f: int = feature_map.temperature_range_f[1]
        self.min_temp_c: int = feature_map.temperature_range_c[0]
        self.max_temp_c: int = feature_map.temperature_range_c[1]
        self.ready_start: bool = False
        self.preheat: bool = False
        self.cook_status: str | None = None
        self.current_temp: int | None = None
        self.cook_set_temp: int | None = None
        self.cook_set_time: int | None = None
        self.cook_last_time: int | None = None
        self.last_timestamp: int | None = None
        self.preheat_set_time: int | None = None
        self.preheat_last_time: int | None = None
        self._temp_unit: str | None = None

    @property
    def is_resumable(self) -> bool:
        """Return if cook is resumable."""
        if self.cook_status in ['cookStop', 'preheatStop']:
            if self.cook_set_time is not None:
                return self.cook_set_time > 0
            if self.preheat_set_time is not None:
                return self.preheat_set_time > 0
        return False

    @property
    def temp_unit(self) -> str | None:
        """Return temperature unit."""
        return self._temp_unit

    @temp_unit.setter
    def temp_unit(self, temp_unit: str) -> None:
        """Set temperature unit."""
        if temp_unit.lower() in ['f', 'fahrenheit', 'fahrenheight']:  # API TYPO
            self._temp_unit = 'fahrenheit'
        elif temp_unit.lower() in ['c', 'celsius']:
            self._temp_unit = 'celsius'
        else:
            msg = f'Invalid temperature unit - {temp_unit}'
            raise ValueError(msg)

    @property
    def preheat_time_remaining(self) -> int:
        """Return preheat time remaining."""
        if self.preheat is False or self.cook_status == 'preheatEnd':
            return 0
        if self.cook_status in ['pullOut', 'preheatStop']:
            if self.preheat_last_time is None:
                return 0
            return int(self.preheat_last_time)
        if self.preheat_last_time is not None and self.last_timestamp is not None:
            return int(
                max(
                    (
                        self.preheat_last_time * 60
                        - (int(time.time()) - self.last_timestamp)
                    )
                    // 60,
                    0,
                )
            )
        return 0

    @property
    def cook_time_remaining(self) -> int:
        """Returns the amount of time remaining if cooking."""
        if self.preheat is True or self.cook_status == 'cookEnd':
            return 0
        if self.cook_status in ['pullOut', 'cookStop']:
            if self.cook_last_time is None:
                return 0
            return int(max(self.cook_last_time, 0))
        if self.cook_last_time is not None and self.last_timestamp is not None:
            return int(
                max(
                    (self.cook_last_time * 60 - (int(time.time()) - self.last_timestamp))
                    // 60,
                    0,
                )
            )
        return 0

    @property
    def remaining_time(self) -> int:
        """Return minutes remaining if cooking/heating."""
        if self.preheat is True:
            return self.preheat_time_remaining
        return self.cook_time_remaining

    @property
    def is_running(self) -> bool:
        """Return if cooking or heating."""
        return bool(self.cook_status in ['cooking', 'heating']) and bool(
            self.remaining_time > 0
        )

    @property
    def is_cooking(self) -> bool:
        """Return if cooking."""
        return self.cook_status == 'cooking' and self.remaining_time > 0

    @property
    def is_heating(self) -> bool:
        """Return if heating."""
        return self.cook_status == 'heating' and self.remaining_time > 0

    def status_request(self, json_cmd: dict) -> None:  # noqa: C901
        """Set status from jsonCmd of API call."""
        self.last_timestamp = None
        if not isinstance(json_cmd, dict):
            return
        self.preheat = False
        preheat = json_cmd.get('preheat')
        cook = json_cmd.get('cookMode')
        if isinstance(preheat, dict):
            self.preheat = True
            if preheat.get('preheatStatus') == 'stop':
                self.cook_status = 'preheatStop'
            elif preheat.get('preheatStatus') == 'heating':
                self.cook_status = 'heating'
                self.last_timestamp = int(time.time())
                self.preheat_set_time = preheat.get(
                    'preheatSetTime', self.preheat_set_time
                )
                if preheat.get('preheatSetTime') is not None:
                    self.preheat_last_time = preheat.get('preheatSetTime')
                self.cook_set_temp = preheat.get('targetTemp', self.cook_set_temp)
                self.cook_set_time = preheat.get('cookSetTime', self.cook_set_time)
                self.cook_last_time = None
            elif preheat.get('preheatStatus') == 'end':
                self.cook_status = 'preheatEnd'
                self.preheat_last_time = 0
        elif isinstance(cook, dict):
            self.clear_preheat()
            if cook.get('cookStatus') == 'stop':
                self.cook_status = 'cookStop'
            elif cook.get('cookStatus') == 'cooking':
                self.cook_status = 'cooking'
                self.last_timestamp = int(time.time())
                self.cook_set_time = cook.get('cookSetTime', self.cook_set_time)
                self.cook_set_temp = cook.get('cookSetTemp', self.cook_set_temp)
                self.current_temp = cook.get('currentTemp', self.current_temp)
                self.temp_unit = cook.get(
                    'tempUnit',
                    self.temp_unit,  # type: ignore[assignment]
                )
            elif cook.get('cookStatus') == 'end':
                self.set_standby()
                self.cook_status = 'cookEnd'

    def clear_preheat(self) -> None:
        """Clear preheat status."""
        self.preheat = False
        self.preheat_set_time = None
        self.preheat_last_time = None

    def set_standby(self) -> None:
        """Clear cooking status."""
        self.cook_status = 'standby'
        self.clear_preheat()
        self.cook_last_time = None
        self.current_temp = None
        self.cook_set_time = None
        self.cook_set_temp = None
        self.last_timestamp = None

    def status_response(self, return_status: dict) -> None:
        """Set status of Air Fryer Based on API Response."""
        self.last_timestamp = None
        self.preheat = False
        self.cook_status = return_status.get('cookStatus')
        if self.cook_status == 'standby':
            self.set_standby()
            return

        #  If drawer is pulled out, set standby if resp does not contain other details
        if self.cook_status == 'pullOut':
            self.last_timestamp = None
            if 'currentTemp' not in return_status or 'tempUnit' not in return_status:
                self.set_standby()
                self.cook_status = 'pullOut'
                return
        if return_status.get('preheatLastTime') is not None or self.cook_status in [
            'heating',
            'preheatStop',
            'preheatEnd',
        ]:
            self.preheat = True

        self.cook_set_time = return_status.get('cookSetTime', self.cook_set_time)
        self.cook_last_time = return_status.get('cookLastTime')
        self.current_temp = return_status.get('curentTemp')
        self.cook_set_temp = return_status.get(
            'targetTemp', return_status.get('cookSetTemp')
        )
        self.temp_unit = return_status.get(
            'tempUnit',
            self.temp_unit,  # type: ignore[assignment]
        )
        self.preheat_set_time = return_status.get('preheatSetTime')
        self.preheat_last_time = return_status.get('preheatLastTime')

        #  Set last_time timestamp if cooking
        if self.cook_status in ['cooking', 'heating']:
            self.last_timestamp = int(time.time())

        if self.cook_status == 'preheatEnd':
            self.preheat_last_time = 0
            self.cook_last_time = None
        if self.cook_status == 'cookEnd':
            self.cook_last_time = 0

        #  If Cooking, clear preheat status
        if self.cook_status in ['cooking', 'cookStop', 'cookEnd']:
            self.clear_preheat()


class VeSyncAirFryer158(VeSyncFryer):
    """Cosori Air Fryer Class.

    Args:
        details (ResponseDeviceDetailsModel): Device details.
        manager (VeSync): Manager class.
        feature_map (DeviceMapTemplate): Device feature map.

    Attributes:
        features (list[str]): List of features.
        state (AirFryer158138State): Air fryer state.
        last_update (int): Last update timestamp.
        refresh_interval (int): Refresh interval in seconds.
        cook_temps (dict[str, list[int]] | None): Cook temperatures.
        pid (str): PID for the device.
        last_response (ResponseInfo): Last response from API call.
        manager (VeSync): Manager object for API calls.
        device_name (str): Name of device.
        device_image (str): URL for device image.
        cid (str): Device ID.
        connection_type (str): Connection type of device.
        device_type (str): Type of device.
        type (str): Type of device.
        uuid (str): UUID of device, not always present.
        config_module (str): Configuration module of device.
        mac_id (str): MAC ID of device.
        current_firm_version (str): Current firmware version of device.
        device_region (str): Region of device. (US, EU, etc.)
        sub_device_no (int): Sub-device number of device.
        product_type (str): Product type of device.
    """

    __slots__ = (
        'cook_temps',
        'last_update',
        'ready_start',
        'refresh_interval',
    )

    def __init__(
        self,
        details: ResponseDeviceDetailsModel,
        manager: VeSync,
        feature_map: AirFryerMap,
    ) -> None:
        """Init the VeSync Air Fryer 158 class."""
        super().__init__(details, manager, feature_map)
        self.features: list[str] = feature_map.features
        self.state: AirFryer158138State = AirFryer158138State(self, details, feature_map)
        self.last_update: int = int(time.time())
        self.refresh_interval = 0
        self.ready_start = False
        self.cook_temps: dict[str, list[int]] | None = None
        if self.config_module not in AIRFRYER_PID_MAP:
            msg = (
                'Report this error as an issue - '
                f'{self.config_module} not found in PID map for {self}'
            )
            raise VeSyncError(msg)
        self.pid = AIRFRYER_PID_MAP[self.config_module]
        self.request_keys = (
            'acceptLanguage',
            'accountID',
            'appVersion',
            'cid',
            'configModule',
            'deviceRegion',
            'phoneBrand',
            'phoneOS',
            'timeZone',
            'token',
            'traceId',
            'userCountryCode',
            'method',
            'debugMode',
            'uuid',
            'pid',
        )

    @deprecated('There is no on/off function for Air Fryers.')
    async def toggle_switch(self, toggle: bool | None = None) -> bool:
        """Turn on or off the air fryer."""
        return toggle if toggle is not None else not self.is_on

    def _build_request(
        self,
        json_cmd: dict | None = None,
        method: str | None = None,
    ) -> dict:
        """Return body of api calls."""
        req_dict = Helpers.get_defaultvalues_attributes(self.request_keys)
        req_dict.update(Helpers.get_manager_attributes(self.manager, self.request_keys))
        req_dict.update(Helpers.get_device_attributes(self, self.request_keys))
        req_dict['method'] = method or 'bypass'
        req_dict['jsonCmd'] = json_cmd or {}
        return req_dict

    def _build_status_body(self, cmd_dict: dict) -> dict:
        """Return body of api calls."""
        body = self._build_request()
        body.update(
            {
                'uuid': self.uuid,
                'configModule': self.config_module,
                'jsonCmd': cmd_dict,
                'pid': self.pid,
                'accountID': self.manager.account_id,
            }
        )
        return body

    @property
    def temp_unit(self) -> str | None:
        """Return temp unit."""
        return self.state.temp_unit

    async def get_details(self) -> None:
        """Get Air Fryer Status and Details."""
        cmd = {'getStatus': 'status'}
        req_body = self._build_request(json_cmd=cmd)
        url = '/cloud/v1/deviceManaged/bypass'
        r_dict, _ = await self.manager.async_call_api(url, 'post', json_object=req_body)
        resp = Helpers.process_dev_response(logger, 'get_details', self, r_dict)
        if resp is None:
            self.state.device_status = DeviceStatus.OFF
            self.state.connection_status = ConnectionStatus.OFFLINE
            return

        return_status = resp.get('result', {}).get('returnStatus')
        if return_status is None:
            LibraryLogger.error_device_response_content(
                logger,
                self,
                'get_details',
                msg='Return status not found in response',
            )
            return
        self.state.status_response(return_status)

    async def check_status(self) -> None:
        """Update status if REFRESH_INTERVAL has passed."""
        seconds_elapsed = int(time.time()) - self.last_update
        logger.debug('Seconds elapsed between updates: %s', seconds_elapsed)
        refresh = False
        if self.refresh_interval is None:
            refresh = bool(seconds_elapsed > REFRESH_INTERVAL)
        elif self.refresh_interval == 0:
            refresh = True
        elif self.refresh_interval > 0:
            refresh = bool(seconds_elapsed > self.refresh_interval)
        if refresh is True:
            logger.debug('Updating status, %s seconds elapsed', seconds_elapsed)
            await self.update()

    async def end(self) -> bool:
        """End the cooking process."""
        await self.check_status()
        if self.state.preheat is False and self.state.cook_status in [
            'cookStop',
            'cooking',
        ]:
            cmd = {'cookMode': {'cookStatus': 'end'}}
        elif self.state.preheat is True and self.state.cook_status in [
            'preheatStop',
            'heating',
        ]:
            cmd = {'preheat': {'cookStatus': 'end'}}
        else:
            logger.debug(
                'Cannot end %s as it is not cooking or preheating', self.device_name
            )
            return False

        status_api = await self._status_api(cmd)
        if status_api is False:
            return False
        self.state.set_standby()
        return True

    async def pause(self) -> bool:
        """Pause the cooking process."""
        await self.check_status()
        if self.state.cook_status not in ['cooking', 'heating']:
            logger.debug(
                'Cannot pause %s as it is not cooking or preheating', self.device_name
            )
            return False
        if self.state.preheat is True:
            cmd = {'preheat': {'preheatStatus': 'stop'}}
        else:
            cmd = {'cookMode': {'cookStatus': 'stop'}}
        status_api = await self._status_api(cmd)
        if status_api is True:
            if self.state.preheat is True:
                self.state.cook_status = 'preheatStop'
            else:
                self.state.cook_status = 'cookStop'
            return True
        return False

    def _validate_temp(self, set_temp: int) -> bool:
        """Temperature validation."""
        if self.state.temp_unit == 'fahrenheit' and (
            set_temp < self.state.min_temp_f or set_temp > self.state.max_temp_f
        ):
            logger.debug('Invalid temperature %s for %s', set_temp, self.device_name)
            return False
        if self.state.temp_unit == 'celsius' and (
            set_temp < self.state.min_temp_c or set_temp > self.state.max_temp_c
        ):
            logger.debug('Invalid temperature %s for %s', set_temp, self.device_name)
            return False
        return True

    async def cook(self, set_temp: int, set_time: int) -> bool:
        """Set cook time and temperature in Minutes."""
        await self.check_status()
        if self._validate_temp(set_temp) is False:
            return False
        return await self._set_cook(set_temp, set_time)

    async def resume(self) -> bool:
        """Resume paused preheat or cook."""
        await self.check_status()
        if self.state.cook_status not in ['preheatStop', 'cookStop']:
            logger.debug('Cannot resume %s as it is not paused', self.device_name)
            return False
        if self.state.preheat is True:
            cmd = {'preheat': {'preheatStatus': 'heating'}}
        else:
            cmd = {'cookMode': {'cookStatus': 'cooking'}}
        status_api = await self._status_api(cmd)
        if status_api is True:
            if self.state.preheat is True:
                self.state.cook_status = 'heating'
            else:
                self.state.cook_status = 'cooking'
            return True
        return False

    async def set_preheat(self, target_temp: int, cook_time: int) -> bool:
        """Set preheat mode with cooking time."""
        await self.check_status()
        if self.state.cook_status not in ['standby', 'cookEnd', 'preheatEnd']:
            logger.debug(
                'Cannot set preheat for %s as it is not in standby', self.device_name
            )
            return False
        if self._validate_temp(target_temp) is False:
            return False
        cmd = self._cmd_api_dict
        cmd['preheatSetTime'] = 5
        cmd['preheatStatus'] = 'heating'
        cmd['targetTemp'] = target_temp
        cmd['cookSetTime'] = cook_time
        json_cmd = {'preheat': cmd}
        return await self._status_api(json_cmd)

    async def cook_from_preheat(self) -> bool:
        """Start Cook when preheat has ended."""
        await self.check_status()
        if self.state.preheat is False or self.state.cook_status != 'preheatEnd':
            logger.debug('Cannot start cook from preheat for %s', self.device_name)
            return False
        return await self._set_cook(status='cooking')

    async def update(self) -> None:
        """Update the device details."""
        await self.get_details()

    @property
    def _cmd_api_base(self) -> dict:
        """Return Base api dictionary for setting status."""
        return {
            'mode': COOK_MODE,
            'accountId': self.manager.account_id,
        }

    @property
    def _cmd_api_dict(self) -> dict:
        """Return API dictionary for setting status."""
        cmd = self._cmd_api_base
        cmd.update(
            {
                'appointmentTs': 0,
                'recipeId': RECIPE_ID,
                'readyStart': self.ready_start,
                'recipeType': RECIPE_TYPE,
                'customRecipe': CUSTOM_RECIPE,
            }
        )
        return cmd

    async def _set_cook(
        self,
        set_temp: int | None = None,
        set_time: int | None = None,
        status: str = 'cooking',
    ) -> bool:
        if set_temp is not None and set_time is not None:
            set_cmd = self._cmd_api_dict

            set_cmd['cookSetTime'] = set_time
            set_cmd['cookSetTemp'] = set_temp
        else:
            set_cmd = self._cmd_api_base
        set_cmd['cookStatus'] = status
        cmd = {'cookMode': set_cmd}
        return await self._status_api(cmd)

    async def _status_api(self, json_cmd: dict) -> bool:
        """Set API status with jsonCmd."""
        body = self._build_status_body(json_cmd)
        url = '/cloud/v1/deviceManaged/bypass'
        r_dict, _ = await self.manager.async_call_api(url, 'post', json_object=body)
        resp = Helpers.process_dev_response(logger, 'set_status', self, r_dict)
        if resp is None:
            return False

        self.last_update = int(time.time())
        self.state.status_request(json_cmd)
        await self.update()
        return True



"""
Outlet Device API Responses

OUTLET variable is a list of device types

DETAILS_RESPONSES variable is a dictionary of responses from the API
for get_details() methods.  The keys are the device types and the
values are the responses.  The responses are tuples of (response, status)

METHOD_RESPONSES variable is a defaultdict of responses from the API. This is
the FunctionResponse variable from the utils module in the tests dir.
The default response is a tuple with the value ({"code": 0, "msg": "success"}, 200).

The values of METHOD_RESPONSES can be a function that takes a single argument or
a static value. The value is checked if callable at runtime and if so, it is called
with the provided argument. If not callable, the value is returned as is.

METHOD_RESPONSES = {
    'DEV_TYPE': defaultdict(
        lambda: ({"code": 0, "msg": "success"}, 200))
    )
}

### For a function to handle the response
def status_response(request_kwargs=None):
    # do work with request_kwargs
    return request_body, 200

METHOD_RESPONSES['DEV_TYPE']['set_status'] = status_response

### To change the default value for a device type

METHOD_RESPONSES['DEVTYPE'].default_factory = lambda: ({"code": 0, "msg": "success"}, 200)

If changing default response for all devices, change the default factory of the import
default dict but make sure to use `deepcopy` to avoid unintended side effects.
"""

from copy import deepcopy
from collections import defaultdict
from pyvesync.const import (
    DeviceStatus,
    ConnectionStatus,
    NightlightModes,
    NightlightStatus,
)
from pyvesync.device_map import outlet_modules
from defaults import (
    FunctionResponses,
    TestDefaults,
    build_bypass_v1_response,
    build_bypass_v2_response,
    FunctionResponsesV1,
    FunctionResponsesV2,
)


OUTLETS = [m.setup_entry for m in outlet_modules]
OUTLETS_NUM = len(OUTLETS)


class OutletDefaults:
    device_status = DeviceStatus.ON
    connection_status = ConnectionStatus.ONLINE
    night_light_status = NightlightStatus.ON
    night_light_mode = NightlightModes.AUTO
    voltage = 120  # volts
    energy = 10  # kilowatt
    power = 20  # kilowatt-hours
    current = 0.5  # amps
    round_7a_voltage = "78000:78000"  # 120 Volts
    round_7a_power = "1000:1000"  # 1 watt


OUTLET_DETAILS: dict[str, dict] = {
    "wifi-switch-1.3": {
        "deviceStatus": OutletDefaults.device_status.value,
        "deviceImg": "https://image.vesync.com/defaultImages/deviceDefaultImages/7aoutlet_240.png",
        "energy": OutletDefaults.energy,
        "activeTime": TestDefaults.active_time,
        "power": OutletDefaults.round_7a_power,
        "voltage": OutletDefaults.round_7a_voltage,
    },
    "ESW03": {
        "activeTime": TestDefaults.active_time,
        "deviceName": "ESW03",
        "deviceStatus": OutletDefaults.device_status.value,
        "power": str(OutletDefaults.power),
        "voltage": str(OutletDefaults.voltage),
        "energy": OutletDefaults.energy,
        "connectionStatus": OutletDefaults.connection_status.value,
    },
    "ESW10-USA": {
        "id": 0,
        "enabled": bool(OutletDefaults.device_status),
    },
    "ESW15-USA": {  # V1
        "activeTime": TestDefaults.active_time,
        "deviceName": "Etekcity 15A WiFi Outlet US/CA",
        "deviceStatus": OutletDefaults.device_status.value,
        "power": OutletDefaults.power,
        "voltage": OutletDefaults.voltage,
        "energy": OutletDefaults.energy,
        "nightLightStatus": OutletDefaults.night_light_status.value,
        "nightLightAutomode": OutletDefaults.night_light_mode.value,
        "nightLightBrightness": 50,
        "deviceImg": "https://image.vesync.com/defaultImages/deviceDefaultImages/15aoutletnightlight_240.png",
        "connectionStatus": OutletDefaults.connection_status.value,
    },
    "ESO15-TB": {
        "activeTime": TestDefaults.active_time,
        "deviceName": "Etekcity Outdoor Plug",
        "deviceStatus": OutletDefaults.device_status.value,
        "power": OutletDefaults.power,
        "voltage": OutletDefaults.voltage,
        "energy": OutletDefaults.energy,
        "subDevices": [
            {
                "subDeviceNo": 1,
                "defaultName": "Socket A",
                "subDeviceName": "Socket A",
                "subDeviceImg": "",
                "subDeviceStatus": OutletDefaults.device_status.value,
            },
            {
                "subDeviceNo": 2,
                "defaultName": "Socket B",
                "subDeviceName": "Socket B",
                "subDeviceImg": "",
                "subDeviceStatus": OutletDefaults.device_status.value,
            },
        ],
        "deviceImg": "",
        "connectionStatus": OutletDefaults.connection_status.value,
    },
    "BSDOG01": {
        "channelNum": 1,
        "powerSwitch_1": 1,
        "realTimeVoltage": OutletDefaults.voltage,
        "realTimePower": OutletDefaults.power,
        "electricalEnergy": OutletDefaults.energy,
        "voltageUpperThreshold": 280,
        "voltageUnderThreshold": 75,
        "protectionStatus": "normal",
        "currentUpperThreshold": 16,
    },
    "WHOGPLUG": {
        "enabled": True,
        "voltage": OutletDefaults.voltage,
        "energy": OutletDefaults.energy,
        "power": OutletDefaults.power,
        "current": OutletDefaults.current,
        "highestVoltage": 240,
        "voltagePtStatus": False,
    },
}


DETAILS_RESPONSES = {
    "wifi-switch-1.3": OUTLET_DETAILS["wifi-switch-1.3"],
    "ESW03": build_bypass_v1_response(result_dict=OUTLET_DETAILS["ESW03"]),
    "ESW10-USA": build_bypass_v2_response(inner_result=OUTLET_DETAILS["ESW10-USA"]),
    "ESW15-USA": build_bypass_v1_response(result_dict=OUTLET_DETAILS["ESW15-USA"]),
    "ESO15-TB": build_bypass_v1_response(result_dict=OUTLET_DETAILS["ESO15-TB"]),
    "BSDOG01": build_bypass_v2_response(inner_result=OUTLET_DETAILS["BSDOG01"]),
    "WHOGPLUG": build_bypass_v2_response(inner_result=OUTLET_DETAILS["WHOGPLUG"]),
}


ENERGY_HISTORY = {
    "traceId": TestDefaults.trace_id,
    "code": 0,
    "msg": "request success",
    "module": None,
    "stacktrace": None,
    "result": {
        "energyConsumptionOfToday": 1,
        "costPerKWH": 0.50,
        "maxEnergy": 0.7,
        "totalEnergy": 2.0,
        "totalMoney": 1.0,
        "averageEnergy": 0.01,
        "averageMoney": 0.0,
        "maxCost": 35.0,
        "energySavingStatus": "off",
        "currency": "USD",
        "energyInfos": [
            {"timestamp": 1742184000025, "energyKWH": 0.0015, "money": 0.0},
            {"timestamp": 1742270400025, "energyKWH": 0.0016, "money": 0.0},
            {"timestamp": 1742356800025, "energyKWH": 0.0017, "money": 0.0},
        ],
    },
}

WHOPLUG_ENERGY_HISTORY_COMMON = {
    "energyInfos": [
        {"timestamp": 1697673600, "energy": 0.06002924054861},
        {"timestamp": 1697587200, "energy": 0.03776630222797},
        {"timestamp": 1697500800, "energy": 0.00002495583819},
        {"timestamp": 1697414400, "energy": 0.00038652237961},
        {"timestamp": 1697328000, "energy": 0},
        {"timestamp": 1697241600, "energy": 0},
        {"timestamp": 1697155200, "energy": 0},
    ],
    "total": 7,
}

WHOPLUG_YEAR_ENERGY_HISTORY = {
    "ELECConsumeList": [
        {"month": "2022-11", "ELECConsume": 0},
        {"month": "2022-12", "ELECConsume": 0},
        {"month": "2023-01", "ELECConsume": 0},
        {"month": "2023-02", "ELECConsume": 0},
        {"month": "2023-03", "ELECConsume": 0},
        {"month": "2023-04", "ELECConsume": 0},
        {"month": "2023-05", "ELECConsume": 0},
        {"month": "2023-06", "ELECConsume": 0},
        {"month": "2023-07", "ELECConsume": 0},
        {"month": "2023-08", "ELECConsume": 0},
        {"month": "2023-09", "ELECConsume": 0},
        {"month": "2023-10", "ELECConsume": 0.0281},
    ]
}

METHOD_RESPONSES = {
    "wifi-switch-1.3": defaultdict(lambda: None),
    "ESW10-USA": deepcopy(FunctionResponsesV2),
    "ESW03": deepcopy(FunctionResponses),
    "ESW15-USA": deepcopy(FunctionResponsesV1),
    "ESO15-TB": deepcopy(FunctionResponsesV1),
    "BSDOG01": deepcopy(FunctionResponsesV2),
    "WHOGPLUG": deepcopy(FunctionResponsesV2),
}

for k in METHOD_RESPONSES:
    if k in ["ESW10-USA"]:
        METHOD_RESPONSES[k]["get_weekly_energy"] = None
        METHOD_RESPONSES[k]["get_monthly_energy"] = None
        METHOD_RESPONSES[k]["get_yearly_energy"] = None
    elif k in ["BSDOG01", "WHOGPLUG"]:
        METHOD_RESPONSES[k]["get_weekly_energy"] = build_bypass_v2_response(
            inner_result=WHOPLUG_ENERGY_HISTORY_COMMON
        )
        METHOD_RESPONSES[k]["get_monthly_energy"] = build_bypass_v2_response(
            inner_result=WHOPLUG_ENERGY_HISTORY_COMMON
        )
        METHOD_RESPONSES[k]["get_yearly_energy"] = build_bypass_v1_response(
            result_dict=WHOPLUG_YEAR_ENERGY_HISTORY
        )
    else:
        METHOD_RESPONSES[k]["get_weekly_energy"] = ENERGY_HISTORY
        METHOD_RESPONSES[k]["get_monthly_energy"] = ENERGY_HISTORY
        METHOD_RESPONSES[k]["get_yearly_energy"] = ENERGY_HISTORY



"""Base class for all VeSync devices."""

from __future__ import annotations

import inspect
import logging
from abc import ABC, abstractmethod
from datetime import UTC
from datetime import datetime as dt
from typing import TYPE_CHECKING, Any, Generic, TypeVar

import orjson

from pyvesync.const import ConnectionStatus, DeviceStatus

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from pyvesync import VeSync
    from pyvesync.device_map import DeviceMapTemplate
    from pyvesync.models.vesync_models import ResponseDeviceDetailsModel
    from pyvesync.utils.errors import ResponseInfo
    from pyvesync.utils.helpers import Timer


VS_TYPE = TypeVar('VS_TYPE', bound='VeSyncBaseDevice')
VS_STATE_T = TypeVar('VS_STATE_T', bound='DeviceState')


class VeSyncBaseDevice(ABC, Generic[VS_STATE_T]):
    """Properties shared across all VeSync devices.

    Abstract Base Class for all VeSync devices. The device class is used solely
    for operational methods and static device properties.

    Parameters:
        details (ResponseDeviceDetailsModel): Device details from API call.
        manager (VeSync): Manager object for API calls.
        feature_map (DeviceMapTemplate): Device configuration map, will be specific
            subclass of DeviceMapTemplate based on device type.

    Attributes:
        state (pyvesync.base_devices.vesyncbasedevice.DeviceState): Device state object
            Each device has a separate state base class in the base_devices module.
        last_response (ResponseInfo): Last response from API call.
        manager (VeSync): Manager object for API calls.
        device_name (str): Name of device.
        device_image (str): URL for device image.
        cid (str): Device ID.
        connection_type (str): Connection type of device.
        device_type (str): Type of device.
        type (str): Type of device.
        uuid (str): UUID of device, not always present.
        config_module (str): Configuration module of device.
        mac_id (str): MAC ID of device.
        current_firm_version (str): Current firmware version of device.
        latest_firm_version (str | None): Latest firmware version of device.
        device_region (str): Region of device. (US, EU, etc.)
        pid (str): Product ID of device, pulled by some devices on update.
        sub_device_no (int): Sub-device number of device.
        product_type (str): Product type of device.
        features (dict): Features of device.

    Methods:
        set_timer: Set timer for device.
        get_timer: Get timer for device from API.
        clear_timer: Clear timer for device from API.
        set_state: Set device state attribute.
        get_state: Get device state attribute.
        update: Update device details.
        display: Print formatted static device info to stdout.
        to_json: Print JSON API string
        to_jsonb: JSON API bytes device details
        to_dict: Return device information as a dictionary.

    Note:
        Device states are found in the `state` attribute in a subclass of DeviceState
        based on the device type. The `DeviceState` subclass is located in device the
        base_devices module.

        The `last_response` attribute is used to store the last response and error
        information from the API call. See the `pyvesync.errors` module for more
        information.
    """

    __slots__ = (
        '__base_exclusions',
        '__weakref__',
        '_exclude_serialization',
        'cid',
        'config_module',
        'connection_type',
        'current_firm_version',
        'device_image',
        'device_name',
        'device_region',
        'device_type',
        'enabled',
        'features',
        'last_response',
        'latest_firm_version',
        'mac_id',
        'manager',
        'pid',
        'product_type',
        'request_keys',
        'state',
        'sub_device_no',
        'type',
        'uuid',
    )

    state: VS_STATE_T

    def __init__(
        self,
        details: ResponseDeviceDetailsModel,
        manager: VeSync,
        feature_map: DeviceMapTemplate,
    ) -> None:
        """Initialize VeSync device base class."""
        self._exclude_serialization: list[str] = []
        self.enabled: bool = True
        self.last_response: ResponseInfo | None = None
        self.manager = manager
        self.device_name: str = details.deviceName
        self.device_image: str | None = details.deviceImg
        self.cid: str = details.cid
        self.connection_type: str | None = details.connectionType
        self.device_type: str = details.deviceType
        self.type: str | None = details.type
        self.uuid: str | None = details.uuid
        self.config_module: str = details.configModule
        self.mac_id: str | None = details.macID
        self.current_firm_version = details.currentFirmVersion
        self.latest_firm_version: str | None = None
        self.device_region: str | None = details.deviceRegion
        self.pid: str | None = None
        self.sub_device_no: int | None = details.subDeviceNo
        # From feature_map
        self.product_type: str = feature_map.product_type
        self.features: list[str] = feature_map.features

    def __eq__(self, other: object) -> bool:
        """Use device CID and sub-device number to test equality."""
        if not isinstance(other, VeSyncBaseDevice):
            return NotImplemented
        return bool(other.cid == self.cid and other.sub_device_no == self.sub_device_no)

    def __hash__(self) -> int:
        """Use CID and sub-device number to make device hash."""
        return hash(self.cid + str(self.sub_device_no))

    def __str__(self) -> str:
        """Use device info for string representation of class."""
        return (
            f'Device Name: {self.device_name}, '
            f'Device Type: {self.device_type}, '
            f'SubDevice No.: {self.sub_device_no}'
        )

    def __repr__(self) -> str:
        """Representation of device details."""
        return (
            f'DevClass: {self.__class__.__name__}, '
            f'Product Type: {self.product_type}, '
            f'Name:{self.device_name}, '
            f'Device No: {self.sub_device_no}, '
            f'CID: {self.cid}'
        )

    @property
    def is_on(self) -> bool:
        """Return true if device is on."""
        return self.state.device_status in [DeviceStatus.ON, DeviceStatus.RUNNING]

    @property
    def firmware_update(self) -> bool:
        """Return True if firmware update available.

        This is going to be updated.
        """
        return self.latest_firm_version != self.current_firm_version

    async def set_timer(self, duration: int, action: str | None = None) -> bool:
        """Set timer for device.

        This may not be implemented for all devices. Please open an issue
        if there is an error.

        Args:
            duration (int): Duration in seconds.
            action (str | None): Action to take when timer expires.

        Returns:
            bool: True if successful, False otherwise.
        """
        del duration
        del action
        logger.debug('Not implemented - set_timer')
        return False

    async def get_timer(self) -> Timer | None:
        """Get timer for device from API and set the `state.Timer` attribute.

        This may not be implemented for all devices. Please open an issue
        if there is an error.

        Note:
            This method may not be implemented for all devices. Please
            open an issue if there is an error.
        """
        logger.debug('Not implemented - get_timer')
        return None

    async def clear_timer(self) -> bool:
        """Clear timer for device from API.

        This may not be implemented for all devices. Please open an issue
        if there is an error.

        Returns:
            bool: True if successful, False otherwise.
        """
        logger.debug('Not implemented - clear_timer')
        return False

    def set_state(self, state_attr: str, stat_value: Any) -> None:  # noqa: ANN401
        """Set device state attribute."""
        setattr(self, state_attr, stat_value)

    def get_state(self, state_attr: str) -> Any:  # noqa: ANN401
        """Get device state attribute."""
        return getattr(self.state, state_attr)

    @abstractmethod
    async def get_details(self) -> None:
        """Get device details.

        This method is defined in each device class to contain
        the logic to pull the device state from the API and update
        the device's `state` attribute. The `update()` method should
        be called to update the device state.
        """

    async def update(self) -> None:
        """Update device details."""
        await self.get_details()

    def display(self, state: bool = True) -> None:
        """Print formatted static device info to stdout.

        Args:
            state (bool): If True, include state in display, defaults to True.

        Example:
        ```
        Device Name:..................Living Room Lamp
        Model:........................ESL100
        Subdevice No:.................0
        Type:.........................wifi
        CID:..........................1234567890abcdef
        ```
        """
        # noinspection SpellCheckingInspection
        display_list = [
            ('Device Name:', self.device_name),
            ('Product Type: ', self.product_type),
            ('Model: ', self.device_type),
            ('Subdevice No: ', str(self.sub_device_no)),
            ('Type: ', self.type),
            ('CID: ', self.cid),
            ('Config Module: ', self.config_module),
            ('Connection Type: ', self.connection_type),
            ('Features', self.features),
            ('Last Response: ', self.last_response),
        ]
        if self.uuid is not None:
            display_list.append(('UUID: ', self.uuid))

        for line in display_list:
            print(f'{line[0]:.<30} {line[1]}')  # noqa: T201
        if state:
            self.state.display()

    def to_json(self, state: bool = True, indent: bool = True) -> str:
        """Print JSON API string for device details.

        Args:
            state (bool): If True, include state in JSON output, defaults to True.
            indent (bool): If True, indent JSON output, defaults to True.

        Returns:
            str: JSON formatted string of device details.
        """
        return self.to_jsonb(state, indent).decode()

    def to_dict(self, state: bool = True) -> dict[str, Any]:
        """Return device information as a dictionary.

        Args:
            state (bool): If True, include state in dictionary, defaults to True.

        Returns:
            dict[str, Any]: Dictionary containing device information.
        """
        device_dict = {
            'device_name': self.device_name,
            'product_type': self.product_type,
            'model': self.device_type,
            'subdevice_no': str(self.sub_device_no),
            'type': self.type,
            'cid': self.cid,
            'features:': self.features,
            'config_module': self.config_module,
            'connection_type': self.connection_type,
            'last_response': self.last_response,
        }
        state_dict = self.state.to_dict() if state else {}
        return device_dict | state_dict

    def to_jsonb(self, state: bool = True, indent: bool = True) -> bytes:
        """JSON API bytes for device details.

        Args:
            state (bool): If True, include state in JSON output, defaults to True.
            indent (bool): If True, indent JSON output, defaults to True.

        Returns:
            bytes: JSON formatted bytes of device details.

        Example:
            This is an example without state.
            ```
            {
                "Device Name": "Living Room Lamp",
                "Model": "ESL100",
                "Subdevice No": "0",
                "Type": "wifi",
                "CID": "1234567890abcdef"
            }
            ```
        """
        return_dict = self.to_dict(state=state)
        if indent:
            return orjson.dumps(
                return_dict,
                option=orjson.OPT_INDENT_2 | orjson.OPT_NON_STR_KEYS,
            )

        return orjson.dumps(return_dict, option=orjson.OPT_NON_STR_KEYS)


class VeSyncBaseToggleDevice(VeSyncBaseDevice, Generic[VS_STATE_T]):
    """Base class for VeSync devices that can be toggled on and off.

    Parameters:
        details (ResponseDeviceDetailsModel): Device details from API call.
        manager (VeSync): Manager object for API calls.
        feature_map (DeviceMapTemplate): Device configuration map, will be specific
            subclass of DeviceMapTemplate based on device type.

    Attributes:
        state (pyvesync.base_devices.vesyncbasedevice.DeviceState): Device state object
            Each device has a separate state base class in the base_devices module.
        last_response (ResponseInfo): Last response from API call.
        manager (VeSync): Manager object for API calls.
        device_name (str): Name of device.
        device_image (str): URL for device image.
        cid (str): Device ID.
        connection_type (str): Connection type of device.
        device_type (str): Type of device.
        type (str): Type of device.
        uuid (str): UUID of device, not always present.
        config_module (str): Configuration module of device.
        mac_id (str): MAC ID of device.
        current_firm_version (str): Current firmware version of device.
        device_region (str): Region of device. (US, EU, etc.)
        pid (str): Product ID of device, pulled by some devices on update.
        sub_device_no (int): Sub-device number of device.
        product_type (str): Product type of device.
        features (dict): Features of device.

    Methods:
        set_timer: Set timer for device.
        get_timer: Get timer for device from API.
        clear_timer: Clear timer for device from API.
        set_state: Set device state attribute.
        get_state: Get device state attribute.
        update: Update device details.
        display: Print formatted static device info to stdout.
        to_json: Print JSON API string
        to_jsonb: JSON API bytes device details
        to_dict: Return device information as a dictionary.
        toggle_switch: Toggle device power on or off.
        turn_on: Turn device on.
        turn_off: Turn device off.

    Note:
        Device states are found in the `state` attribute in a subclass of DeviceState
        based on the device type. The `DeviceState` subclass is located in device the
        base_devices module.

        The `last_response` attribute is used to store the last response and error
        information from the API call. See the `pyvesync.errors` module for more
        information.
    """

    __slots__ = ()

    @abstractmethod
    async def toggle_switch(self, toggle: bool | None = None) -> bool:
        """Toggle device power on or off.

        Args:
            toggle (bool | None): True to turn on, False to turn off, None to toggle.

        Returns:
            bool: True if successful, False otherwise.
        """

    async def turn_on(self) -> bool:
        """Turn device on."""
        return await self.toggle_switch(True)

    async def turn_off(self) -> bool:
        """Turn device off."""
        return await self.toggle_switch(False)


class DeviceState:
    """Base dataclass to hold device state.

    Args:
        device (VeSyncBaseDevice): Device object.
        details (ResponseDeviceDetailsModel): Device details from API call.
        feature_map (DeviceMapTemplate): Device configuration map, will be specific
            subclass of DeviceMapTemplate based on device type.

    Attributes:
        active_time (int): Active time of device, defaults to None.
        connection_status (str): Connection status of device.
        device (VeSyncBaseDevice): Device object.
        device_status (str): Device status.
        features (dict): Features of device.
        last_update_ts (int): Last update timestamp in UTC, defaults to None.

    Methods:
        update_ts: Update last update timestamp.
        to_dict: Dump state to JSON.
        to_json: Dump state to JSON string.
        to_jsonb: Dump state to JSON bytes.
        as_tuple: Convert state to tuple of (name, value) tuples.

    Note:
        This cannot be instantiated directly. It should be inherited by the state class
        of a specific product type.
    """

    __slots__ = (
        '__base_exclusions',
        '_exclude_serialization',
        'active_time',
        'connection_status',
        'device',
        'device_status',
        'features',
        'last_update_ts',
        'timer',
    )

    def __init__(
        self,
        device: VeSyncBaseDevice,
        details: ResponseDeviceDetailsModel,
        feature_map: DeviceMapTemplate,
    ) -> None:
        """Initialize device state."""
        self.__base_exclusions: list[str] = ['manager', 'device', 'state']
        self._exclude_serialization: list[str] = []
        self.device = device
        self.device_status: str = details.deviceStatus or DeviceStatus.UNKNOWN
        self.connection_status: str = details.connectionStatus or ConnectionStatus.UNKNOWN
        self.features = feature_map.features
        self.last_update_ts: int | None = None
        self.active_time: int | None = None
        self.timer: Timer | None = None

    def __str__(self) -> str:
        """Return device state as string."""
        return (
            f'{self.device.device_name}, {self.device.product_type}, '
            f'Device Status:{self.device_status}, '
            f'Connection Status: {self.connection_status}'
        )

    def __repr__(self) -> str:
        """Return device state as string."""
        return (
            f'{self.device.device_name}, Device Status:{self.device_status}, '
            f'Connection Status: {self.connection_status}'
        )

    def update_ts(self) -> None:
        """Update last update timestamp as UTC timestamp."""
        self.last_update_ts = int(dt.now(tz=UTC).timestamp())

    @staticmethod
    def __predicate(attr: Any) -> bool:  # noqa: ANN401
        """Check if attribute should be serialized."""
        return (
            callable(attr)
            or inspect.ismethod(attr)
            or inspect.isbuiltin(attr)
            or inspect.isfunction(attr)
            or inspect.isroutine(attr)
            or inspect.isclass(attr)
        )

    def _serialize(self) -> dict[str, Any]:
        """Get dictionary of state attributes."""
        state_dict: dict[str, Any] = {}
        for attr_name, attr_value in inspect.getmembers(
            self, lambda a: not self.__predicate(a)
        ):
            if attr_name.startswith('_') or attr_name in [
                *self.__base_exclusions,
                *self._exclude_serialization,
            ]:
                continue
            state_dict[attr_name] = attr_value
        return state_dict

    def to_json(self, indent: bool = False) -> str:
        """Dump state to JSON string.

        Args:
            indent (bool): If True, indent JSON output, defaults to False.

        Returns:
            str: JSON formatted string of device state.
        """
        return self.to_jsonb(indent=indent).decode()

    def to_jsonb(self, indent: bool = False) -> bytes:
        """Convert state to JSON bytes."""
        if indent:
            return orjson.dumps(
                self._serialize(), option=orjson.OPT_NON_STR_KEYS | orjson.OPT_INDENT_2
            )
        return orjson.dumps(self._serialize(), option=orjson.OPT_NON_STR_KEYS)

    def to_dict(self) -> dict[str, Any]:
        """Convert state to dictionary."""
        return self._serialize()

    def as_tuple(self) -> tuple[tuple[str, Any], ...]:
        """Convert state to tuple of (name, value) tuples."""
        return tuple((k, v) for k, v in self._serialize().items())

    def display(self) -> None:
        """Print formatted state to stdout."""
        for name, val in self._serialize().items():
            print(f'{name:.<30} {val}')  # noqa: T201



"""Contains base test cases for mocking devices and API calls."""

from __future__ import annotations
from typing import TYPE_CHECKING
import asyncio
import pytest
from unittest.mock import MagicMock, patch

from pyvesync import VeSync
from pyvesync.models.vesync_models import ResponseDeviceDetailsModel
from defaults import TestDefaults, API_DEFAULTS
from call_json import ALL_DEVICE_MAP_DICT, DeviceList

if TYPE_CHECKING:
    from pyvesync.base_devices import VeSyncBaseDevice


class TestApiFunc:
    """Test case to instantiate vs object and patch the session object."""
    default_endpoint = '/endpoint'

    @pytest.fixture(autouse=True, scope='function')
    def setup(self, caplog):
        """Fixture to instantiate VeSync object, start logging and start Mock.

        This fixture mocks the ClientSession object directly.

        Yields:
            self: Class instance with mocked session object to test response handling.
        """
        self.mock_api_call = patch("pyvesync.vesync.ClientSession")
        self.caplog = caplog
        self.mock_api = self.mock_api_call.start()
        self.loop = asyncio.new_event_loop()
        self.mock = MagicMock()
        self.manager = VeSync(API_DEFAULTS['EMAIL'], API_DEFAULTS['PASSWORD'])
        self.manager.enabled = True
        self.manager.redact = False
        self.manager.auth._token = TestDefaults.token
        self.manager.auth._account_id = TestDefaults.account_id
        yield
        self.mock.stop()
        self.loop.stop()

    async def run_coro(self, coro):
        """Run a coroutine in the event loop."""
        return await coro

    def run_in_loop(self, func, *args, **kwargs):
        """Run a function in the event loop."""
        return self.loop.run_until_complete(self.run_coro(func(*args, **kwargs)))


class TestBase:
    """Base class for tests with the call_api() method mocked.

    Contains instantiated VeSync object and mocked API call for call_api() function.

    Attributes:
        mock_api (Mock): Mock for call_api() function.
        manager (VeSync): Instantiated VeSync object that is logged in.
        caplog (LogCaptureFixture): Pytest fixture for capturing logs.
    """
    overwrite = False
    write_api = False

    @pytest.fixture(autouse=True, scope='function')
    def setup(self, caplog):
        """Fixture to instantiate VeSync object, start logging and start Mock.

        Attributes
        ----------
        self.mock_api : Mock
        self.manager : VeSync
        self.caplog : LogCaptureFixture

        Yields
        ------
        Class instance with mocked call_api() function and VeSync object
        """
        self.loop = asyncio.new_event_loop()
        self.mock_api_call = patch('pyvesync.vesync.VeSync.async_call_api')
        self.caplog = caplog
        self.mock_api = self.mock_api_call.start()
        self.mock_api.return_value.ok = True
        self.manager = VeSync(TestDefaults.email, TestDefaults.password)
        self.manager.redact = False
        self.manager.time_zone = TestDefaults.time_zone
        self.manager.enabled = True
        self.manager.auth._token = TestDefaults.token
        self.manager.auth._account_id = TestDefaults.account_id
        yield
        self.mock_api_call.stop()

    async def run_coro(self, coro):
        """Run a coroutine in the event loop."""
        return await coro

    def run_in_loop(self, func, *args, **kwargs) -> VeSyncBaseDevice | None:
        """Run a function in the event loop."""
        return self.loop.run_until_complete(self.run_coro(func(*args, **kwargs)))

    def get_device(self, product_type: str, setup_entry: str) -> VeSyncBaseDevice:
        """Get device from device details dict.

        Args:
            device_details (dict): Device details dictionary from call_json module.

        Returns:
            Device object from VeSync.devices
        """
        if len(self.manager.devices) > 0:
            self.manager.devices.clear()
        device_map = ALL_DEVICE_MAP_DICT[setup_entry]
        device_config = DeviceList.device_list_item(device_map)
        self.manager.devices.add_device_from_model(
            ResponseDeviceDetailsModel.from_dict(device_config), self.manager
        )
        device_list = getattr(self.manager.devices, product_type)
        assert len(device_list) == 1, f"Could not instantiate {product_type} device."
        return device_list[0]



"""Testing script for pyvesync library.

This script tests the functionality of the pyvesync library by
interacting with VeSync devices. It can be run from the command line or
in VS code. It supports testing devices, timers, and logging to a file.

Usage:
    python testing_scripts/vs_console_script.py \
            --email <your_email> \
            --password <your_password> \
            [--test_devices] \
            [--test_timers] \
            [--output-file <output_file>] \
            [--test_dev_type <device_type>]

By default, it will only log in, pull the device list, and update all devices. Using
the `--test_devices` flag will run device specific methods on all devices. The
`--test_timers` flag will run timer methods on all devices. To test a specific device
type, use the `--test_dev_type` flag with one of the following values: "bulbs",
"switches", "outlets", "humidifiers", "air_purifiers", "fans". If no device type
is specified, it will test all devices.

Example:
    The script can be debugged in an IDE by setting the USERNAME, PASSWORD and
    other arguments at the top of the file. This injects the values into the
    command line arguments. Command line arguments will override these values.
"""

from __future__ import annotations

import argparse
import asyncio
import logging
import random
import sys
from pathlib import Path
from typing import TYPE_CHECKING, Literal

from pyvesync import VeSync
from pyvesync.const import PurifierModes

if TYPE_CHECKING:
    from pyvesync.base_devices import VeSyncBaseToggleDevice
    from pyvesync.device_container import DeviceContainer

_T_DEVS = Literal['bulbs', 'switches', 'outlets', 'humidifiers', 'air_purifiers', 'fans']

# Manually configure script arguments
USERNAME: str | None = None
PASSWORD: str | None = None
REGION: str | None = None
TEST_DEVICES: bool | None = None
TEST_TIMERS: bool | None = None
OUTPUT_FILE: str | None = None
TEST_DEV_TYPE: _T_DEVS | None = None
# Set to a specific device type to test, e.g., "vesyncbulb", "vesyncswitch", etc.


logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)
logger.info('Starting VeSync device test...')
stream_handler = logging.StreamHandler(sys.stdout)
stream_handler.setLevel(logging.INFO)
formatter = logging.Formatter(
    fmt='%(asctime)s - %(levelname)s - %(name)s - %(message)s',
    datefmt='%Y-%m-%d %H:%M:%S',
)
stream_handler.setFormatter(formatter)
logger.addHandler(stream_handler)
logger.propagate = False

vs_logger = logging.getLogger('pyvesync')
vs_logger.setLevel(logging.DEBUG)

_SEP = '-' * 18
# Logging messages
_LOG_DEVICE_START = _SEP + ' Testing %s %s ' + _SEP


async def _random_await() -> None:
    await asyncio.sleep(random.randrange(10, 30) / 10)  # noqa: S311
    logger.info('Random await finished')


async def vesync_test(  # noqa: PLR0913
    username: str,
    password: str,
    country_code: str,
    output_file: str,
    test_devices: bool = False,
    test_timers: bool = False,
    test_dev_type: _T_DEVS | None = None,
) -> None:
    """Main function to test VeSync devices.

    Note:
        This will attempt to return the device to the original state,
        but it is not guaranteed. Timers may be cleared as well.

    Args:
        username (str): VeSync account email.
        password (str): VeSync account password.
        country_code (str): VeSync account region.
        output_file (str): Path to the output file for logging.
        test_devices (bool): If True, run tests on devices.
        test_timers (bool): If True, run tests on timers.
        test_dev_type (_T_DEVS | None): Specific device type to test.

    """
    output_path = Path(output_file).resolve()
    if not output_path.parent.exists():
        logger.error('Output directory %s does not exist.', output_path.parent)
        return
    file_handler = logging.FileHandler(output_path)
    logger.addHandler(file_handler)

    # Instantiate VeSync object
    async with VeSync(username, password, country_code or 'US') as vs:
        vs.redact = True  # Redact sensitive information in logs
        vs.log_to_file(str(output_path), stdout=False)  # Log to specified file
        logger.info('VeSync instance created, logging to %s', output_path)
        # Login to VeSync account and check for success
        logger.info('Logging in to VeSync account...')
        await vs.login()
        logger.info('Login successful, pulling device list...')
        await _random_await()

        # Pull device list and update all devices
        await vs.update()
        await _random_await()
        logger.info('Device list pulled successfully.')

        # Run tests for outlets
        if vs.devices.outlets and (test_devices is True or test_dev_type == 'outlets'):
            await outlets(vs, update=False)

        # Run tests for air purifiers
        if vs.devices.air_purifiers and (
            test_devices is True or test_dev_type == 'air_purifiers'
        ):
            await air_purifiers(vs, update=False)

        # Run tests for humidifiers
        if vs.devices.humidifiers and (
            test_devices is True or test_dev_type == 'humidifiers'
        ):
            await humidifiers(vs, update=False)

        # Run tests for bulbs
        if vs.devices.bulbs and (test_devices is True or test_dev_type == 'bulbs'):
            await bulbs(vs, update=False)

        # Run tests for switches
        if vs.devices.switches and (test_devices is True or test_dev_type == 'switches'):
            await switches(vs, update=False)

        # Test timers if requested
        if test_timers is True:
            await device_timers(vs.devices, update=False)

    await vs.__aexit__(None, None, None)  # Clean up VeSync instance
    logger.info('Finished testing VeSync devices, results logged to %s', output_path)


async def common_tests(
    device: VeSyncBaseToggleDevice, update: bool = False
) -> bool | None:
    """Run common tests for a VeSync device."""
    logger.info(_LOG_DEVICE_START, device.device_type, device.device_name)
    logger.info(device)

    if update is True:
        await device.update()
        await _random_await()

    if device.state.connection_status == 'offline':
        logger.info('%s is offline, skipping', device.device_name)
        return None

    logger.info('Current device state: %s', device.state.to_json(indent=True))

    if device.state.device_status == 'off':
        logger.info('Initial state is off, turning on device')
        await device.turn_on()
        logger.info(device.state.device_status)
        await _random_await()
        return False

    logger.info('Initial state is on, turning off device')
    await device.turn_off()
    logger.info(device.state.device_status)
    await _random_await()

    logger.info('Turning on device')
    await device.turn_on()
    logger.info(device.state.device_status)
    await _random_await()
    return True


async def device_timers(dev_list: DeviceContainer, update: bool = False) -> None:
    """Test timers for all devices in the device list."""
    # Get dev types from device list to not repeat tests
    dev_types = [dev.device_type for dev in dev_list]
    dev_set = set(dev_types)

    for dev in dev_list:
        logger.info(
            '%s Testing timer for %s %s %s', _SEP, dev.device_type, dev.device_name, _SEP
        )

        if dev.device_type not in dev_set:
            continue
        dev_set.remove(dev.device_type)

        if update is True:
            await dev.update()
            await _random_await()

        await dev.get_timer()
        logger.info('Current timer state: %s', dev.state.timer)
        await _random_await()

        if dev.state.timer is None:
            logger.info('Setting timer for 1 hour to turn on device')
            await dev.set_timer(3600, 'on')
            logger.info('Current timer state: %s', dev.state.timer)
            await _random_await()

        logger.info('Updating device state after setting timer')
        await dev.update()
        logger.info('Current timer state: %s', dev.state.timer)
        await _random_await()

        logger.info('Getting current active timer via API')
        await dev.get_timer()
        logger.info('Current timer state: %s', dev.state.timer)
        await _random_await()

        logger.info('Clearing timer')
        await dev.clear_timer()
        logger.info('Current timer state: %s', dev.state.timer)
        await _random_await()

        logger.info(
            '%s Finished testing timer for %s %s %s',
            _SEP,
            dev.device_type,
            dev.device_name,
            _SEP,
        )


async def humidifiers(manager: VeSync, update: bool = False) -> None:
    """Test humidifiers in the VeSync device manager."""
    logger.info('%s Testing humidifiers %s', _SEP, _SEP)
    dev_types = set()
    for dev in manager.devices.humidifiers:
        # Ensure the same device type is not tested multiple times
        if dev.device_type in dev_types:
            continue
        dev_types.add(dev.device_type)

        initial_on = await common_tests(dev, update)
        if initial_on is None:
            continue

        logger.info(dev.state.to_json(indent=True))
        initial_state = dev.state.to_dict()
        logger.debug('Initial state: %s', dev.state.to_json(indent=True))

        logger.info('Setting state to auto mode')
        await dev.set_auto_mode()
        logger.debug(dev.state.mode)
        await _random_await()

        logger.info('Setting state to manual mode')
        await dev.set_manual_mode()
        logger.debug(dev.state.mode)
        await _random_await()

        logger.info('Setting mist level to 2')
        await dev.set_mist_level(2)
        logger.debug('mist_level: %s', dev.state.mist_level)
        logger.debug('mist_virtual_level: %s', dev.state.mist_virtual_level)
        await _random_await()

        logger.info('Setting sleep mode')
        await dev.set_sleep_mode()
        logger.debug(dev.state.mode)
        await _random_await()

        logger.info('Setting target humidity to 50%')
        await dev.set_humidity(50)
        logger.debug(dev.state.target_humidity)
        await _random_await()

        if initial_on is True:
            logger.info('Returning to initial state')
            await dev.set_mist_level(initial_state['mist_level'])
            await dev.set_mode(initial_state['mode'])
            await dev.set_humidity(initial_state['target_humidity'])
        else:
            await dev.turn_off()
        await _random_await()


async def bulbs(manager: VeSync, update: bool = False) -> None:  # noqa: C901
    """Test bulbs in the VeSync device manager."""
    logger.info('%s Testing bulbs %s', _SEP, _SEP)
    dev_types = set()
    for dev in manager.devices.bulbs:
        if dev.device_type in dev_types:
            continue
        dev_types.add(dev.device_type)

        initial_on = await common_tests(dev, update)
        if initial_on is None:
            continue

        initial_state = dev.state.to_dict()
        logger.info(dev.state.to_json(indent=True))

        # Print the current state of the device
        if dev.supports_brightness:
            logger.info('Setting brightness to 30%')
            await dev.set_brightness(30)
            logger.debug('Brightness: %s', dev.state.brightness)
            await _random_await()

        if dev.supports_color_temp:
            logger.info('Setting color temperature to 10')
            await dev.set_color_temp(10)
            logger.debug('Color Temperature: %s', dev.state.color_temp)
            await _random_await()

        if dev.supports_multicolor:
            logger.info('Setting RGB color to (255, 40, 30)')
            await dev.set_rgb(255, 40, 30)
            logger.debug('Color: %s', dev.state.rgb)
            await _random_await()

        logger.info('Returning to initial state')
        if initial_on is False:
            await dev.turn_off()
            await _random_await()
            return

        if initial_state['color_temp'] is not None:
            await dev.set_color_temp(initial_state['color_temp'])
        if initial_state['brightness'] is not None:
            await dev.set_brightness(initial_state['brightness'])
        if initial_state['rgb'] is not None:
            await dev.set_rgb(
                initial_state['rgb'][0],
                initial_state['rgb'][1],
                initial_state['rgb'][2],
            )

        await _random_await()


async def switches(manager: VeSync, update: bool = False) -> None:  # noqa: C901, PLR0915
    """Test switches in the VeSync device manager."""
    logger.info('%s Testing switches %s', _SEP, _SEP)
    dev_types = set()
    for dev in manager.devices.switches:
        if dev.device_type in dev_types:
            continue
        dev_types.add(dev.device_type)

        initial_on = await common_tests(dev, update)
        if initial_on is None:
            continue

        logger.debug(dev.state.to_json(indent=True))
        initial_state = dev.state.to_dict()

        if dev.supports_dimmable:
            logger.debug('Setting brightness to 100%')
            await dev.set_brightness(100)
            logger.debug(dev.state.brightness)
            await _random_await()

        if dev.supports_indicator_light:
            if initial_state['indicator_status'] == 'on':
                logger.debug('Initial state is on, turning off indicator light')
                await dev.turn_off_indicator_light()
                logger.debug('indicator_status: %s', dev.state.indicator_status)
                await _random_await()
                await dev.turn_on_indicator_light()
                logger.debug('indicator_status: %s', dev.state.indicator_status)
                await _random_await()
            else:
                logger.debug('Initial state is off, turning on indicator light')
                await dev.turn_on_indicator_light()
                logger.debug('indicator_status: %s', dev.state.indicator_status)
                await _random_await()
                await dev.turn_off_indicator_light()
                logger.debug('indicator_status: %s', dev.state.indicator_status)
                await _random_await()

        if dev.supports_backlight_color:
            if initial_state['backlight_status'] == 'on':
                logger.debug('Initial state is on, turning off RGB backlight')
                await dev.turn_off_rgb_backlight()
                logger.debug('backlight_status: %s', dev.state.backlight_status)
                await _random_await()
            logger.debug('Turrning on RGB backlight')
            await dev.turn_on_rgb_backlight()
            logger.debug('backlight_status: %s', dev.state.backlight_status)
            await _random_await()

            logger.debug('Setting backlight color to (50, 30, 15)')
            await dev.set_backlight_color(50, 30, 15)
            logger.debug('backlight_color: %s', dev.state.backlight_color)
            await _random_await()

            if initial_state['backlight_status'] == 'off':
                await dev.turn_off_rgb_backlight()
                await _random_await()
            else:
                await dev.set_backlight_color(
                    initial_state['backlight_color'][0],
                    initial_state['backlight_color'][1],
                    initial_state['backlight_color'][2],
                )
            await _random_await()

        if initial_state['brightness'] is not None:
            await dev.set_brightness(initial_state['brightness'])
            await _random_await()


async def air_purifiers(manager: VeSync, update: bool = False) -> None:  # noqa: C901, PLR0915
    """Test air purifiers in the VeSync device manager."""
    logger.info('%s Testing air purifiers %s', _SEP, _SEP)
    dev_types = set()
    for dev in manager.devices.air_purifiers:
        if dev.device_type in dev_types:
            continue
        dev_types.add(dev.device_type)

        initial_on = await common_tests(dev, update)
        if initial_on is None:
            continue

        logger.info(dev.state.to_json(indent=True))
        initial_state = dev.state.to_dict()

        if initial_state['display_set_state'] is True:
            logger.info('Turning off display')
            await dev.turn_off_display()
            logger.info('display_set_state: %s', dev.state.display_set_status)
            logger.info('display_status: %s', dev.state.display_status)
            await _random_await()
            logger.info('Turning on display')
            await dev.turn_on_display()
            logger.info('display_set_state: %s', dev.state.display_set_status)
            logger.info('display_status: %s', dev.state.display_status)
            await _random_await()
        else:
            logger.info('Display is off, turning on display')
            await dev.turn_on_display()
            logger.info('display_set_state: %s', dev.state.display_set_status)
            logger.info('display_status: %s', dev.state.display_status)
            await _random_await()
            logger.info('Turning off display')
            await dev.turn_off_display()
            logger.info('display_set_state: %s', dev.state.display_set_status)
            logger.info('display_status: %s', dev.state.display_status)
            await _random_await()

        if PurifierModes.AUTO in dev.modes:
            logger.info('Setting auto mode')
            await dev.set_auto_mode()
            logger.info('mode: %s', dev.state.mode)
            await _random_await()
        if PurifierModes.SLEEP in dev.modes:
            logger.info('Setting sleep mode')
            await dev.set_sleep_mode()
            logger.info('mode: %s', dev.state.mode)
            await _random_await()
        if PurifierModes.MANUAL in dev.modes:
            logger.info('Setting manual mode')
            await dev.set_manual_mode()
            logger.info('mode: %s', dev.state.mode)
            await _random_await()

        logger.info('Setting fan speed to 1')
        await dev.set_fan_speed(1)
        logger.info('fan_level: %s', dev.state.fan_level)
        await dev.set_fan_speed(initial_state['fan_level'])

        if initial_state['child_lock'] is True:
            logger.info('Turning off child lock')
            await dev.turn_off_child_lock()
            logger.info('child_lock: %s', dev.state.child_lock)
            await _random_await()
            await dev.turn_on_child_lock()
            await _random_await()
        else:
            logger.info('Turning on child lock')
            await dev.turn_on_child_lock()
            logger.info('child_lock: %s', dev.state.child_lock)
            await _random_await()
            await dev.turn_off_child_lock()
            await _random_await()

        if initial_on is False:
            logger.info('Turning off device')
            await dev.turn_off()
            await _random_await()
            return
        if initial_state['mode'] is not PurifierModes.MANUAL:
            await dev.set_mode(initial_state['mode'])
            await _random_await()


async def outlets(manager: VeSync, update: bool = False) -> None:
    """Test outlets in the VeSync device manager."""
    logger.info('%s Testing outlets %s', _SEP, _SEP)
    dev_types = set()
    for dev in manager.devices.outlets:
        if dev.device_type in dev_types:
            continue
        dev_types.add(dev.device_type)
        initial_on = await common_tests(dev, update)
        if initial_on is None:
            continue
        logger.info(dev.state.to_json(indent=True))
        initial_state = dev.state.to_dict()

        if dev.supports_nightlight:
            if initial_state['nightlight_status'] == 'on':
                logger.info('Turning off nightlight')
                await dev.turn_off_nightlight()
                logger.info('nightlight_status: %s', dev.state.nightlight_status)
                await _random_await()
                await dev.turn_on_nightlight()
            else:
                logger.info('Turning on nightlight')
                await dev.turn_on_nightlight()
                logger.info('nightlight_status: %s', dev.state.nightlight_status)
                await _random_await()
                await dev.turn_off_nightlight()

        if dev.supports_energy:
            logger.info('Getting energy data')
            await dev.get_monthly_energy()
            await dev.get_weekly_energy()
            await dev.get_yearly_energy()
            logger.info('Energy usage: %s', dev.state.energy)
            logger.info('Power: %s', dev.state.power)
            logger.info('Voltage: %s', dev.state.voltage)
            if dev.state.monthly_history:
                logger.info('Monthly history: %s', dev.state.monthly_history.to_json())
            if dev.state.weekly_history:
                logger.info('Weekly history: %s', dev.state.weekly_history.to_json())


if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        description='Test pyvesync library and devices.',
        epilog='Run script with --help for more information.',
    )
    parser.add_argument('--email', default=USERNAME, help='Email for VeSync account')
    parser.add_argument(
        '--password', default=PASSWORD, help='Password for VeSync account'
    )
    parser.add_argument(
        '--country-code',
        default=REGION,
        help='Country of the VeSync account in ISO 3166 Alpha-2 format',
    )
    parser.add_argument(
        '--test_devices',
        action='store_true',
        default=TEST_DEVICES,
        help='Include devices in test, default: False',
    )
    parser.add_argument(
        '--test_timers',
        action='store_true',
        default=TEST_TIMERS,
        help='Run tests on timers, default: False',
    )
    parser.add_argument(
        '--output-file',
        default='vesync.log',
        help='Relative or absolute path to output file',
    )
    parser.add_argument(
        '--test_dev_type',
        choices=['bulbs', 'switches', 'outlets', 'humidifiers', 'air_purifiers', 'fans'],
        default=TEST_DEV_TYPE,
        help="Specific device type to test, e.g., 'bulbs', 'switches', etc.",
    )
    args = parser.parse_args()

    if args.email is None or args.password is None:
        logger.error(
            'Username and password must be provided'
            'via command line arguments or script variables.'
        )
        sys.exit(1)

    asyncio.run(
        vesync_test(
            args.email,
            args.password,
            args.country_code,
            args.output_file,
            args.test_devices,
            args.test_timers,
            args.test_dev_type,
        )
    )



"""Levoit Air Purifier tests."""
import orjson
from pyvesync.devices.vesyncpurifier import VeSyncAirBypass
from pyvesync.devices.vesyncpurifier import VeSyncAir131
from pyvesync.utils.helpers import Helpers as Helpers
import call_json
import call_json_fans
from base_test_cases import TestBase
from defaults import TestDefaults

LVPUR131S = 'LV-PUR131S'
CORE200S = 'Core200S'

DEV_LIST_DETAIL = call_json.DeviceList.device_list_item(LVPUR131S)
CORE200S_DETAIL = call_json.DeviceList.device_list_item(CORE200S)

CORRECT_LIST = call_json.DeviceList.device_list_response(LVPUR131S)

CORRECT_DETAILS = call_json_fans.DETAILS_RESPONSES[LVPUR131S]

BAD_LIST = call_json.DETAILS_BADCODE


class TestVesyncAirPurifier(TestBase):
    """Air purifier tests."""

    def test_airpur_conf(self):
        """Tests that 15A Outlet is instantiated properly."""
        self.mock_api.return_value = CORRECT_LIST
        self.run_in_loop(self.manager.get_devices)
        fans = self.manager.fans
        assert len(fans) == 1
        fan = fans[0]
        assert isinstance(fan, VeSyncAir131)
        assert fan.device_name == TestDefaults.name(LVPUR131S)
        assert fan.device_type == LVPUR131S
        assert fan.cid == TestDefaults.cid(LVPUR131S)
        assert fan.uuid == TestDefaults.uuid(LVPUR131S)

    def test_airpur_details(self):
        """Test Air Purifier get_details()."""
        resp_dict, status = CORRECT_DETAILS
        self.mock_api.return_value = orjson.dumps(resp_dict), status
        fan = VeSyncAir131(DEV_LIST_DETAIL, self.manager)
        self.run_in_loop(fan.get_details)
        dev_details = fan.details
        assert fan.device_status == 'on'
        assert isinstance(dev_details, dict)
        assert dev_details['active_time'] == 1
        assert fan.filter_life == call_json_fans.FanDefaults.filter_life
        assert dev_details['screen_status'] == TestDefaults.str_toggle
        assert fan.mode == 'manual'
        assert dev_details['level'] == call_json_fans.FanDefaults.fan_level
        assert fan.fan_level == call_json_fans.FanDefaults.fan_level
        assert dev_details['air_quality'] == 'excellent'
        assert fan.air_quality == 'excellent'

    def test_airpur_details_fail(self):
        """Test Air Purifier get_details with Code>0."""
        resp_dict, status = BAD_LIST
        self.mock_api.return_value = orjson.dumps(resp_dict), status
        fan = VeSyncAir131(DEV_LIST_DETAIL, self.manager)
        self.run_in_loop(fan.get_details)
        assert len(self.caplog.records) == 1
        assert 'details' in self.caplog.text

    def test_airpur_onoff(self):
        """Test Air Purifier Device On/Off Methods."""
        self.mock_api.return_value = (orjson.dumps({'code': 0}), 200)
        fan = VeSyncAir131(DEV_LIST_DETAIL, self.manager)
        head = Helpers.req_legacy_headers(self.manager)
        body = Helpers.req_body(self.manager, 'devicestatus')
        fan.device_status = 'off'
        body['status'] = 'on'
        body['uuid'] = fan.uuid
        on = self.run_in_loop(fan.turn_on)
        self.mock_api.assert_called_with(
            '/131airPurifier/v1/device/deviceStatus', 'put',
            json_object=body, headers=head)
        call_args = self.mock_api.call_args_list[0][0]
        assert call_args[0] == '/131airPurifier/v1/device/deviceStatus'
        assert call_args[1] == 'put'
        assert on
        fan.device_status = 'on'
        off = self.run_in_loop(fan.turn_off)
        body['status'] = 'off'
        self.mock_api.assert_called_with(
            '/131airPurifier/v1/device/deviceStatus', 'put',
            json_object=body, headers=head)
        assert off

    def test_airpur_onoff_fail(self):
        """Test Air Purifier On/Off Fail with Code>0."""
        self.mock_api.return_value = (orjson.dumps({'code': 1}), 400)
        vsfan = VeSyncAir131(DEV_LIST_DETAIL, self.manager)
        assert not self.run_in_loop(vsfan.turn_on)
        assert not self.run_in_loop(vsfan.turn_off)

    def test_airpur_fanspeed(self):
        """Test changing fan speed of."""
        self.mock_api.return_value = (orjson.dumps({'code': 0}), 200)
        fan = VeSyncAir131(DEV_LIST_DETAIL, self.manager)
        fan.mode = 'manual'
        fan.details['level'] = 1
        b = self.run_in_loop(fan.set_fan_speed)
        assert fan.fan_level == 2
        b = self.run_in_loop(fan.set_fan_speed)
        assert fan.fan_level == 3
        b = self.run_in_loop(fan.set_fan_speed)
        assert fan.fan_level == 1
        assert b
        b = self.run_in_loop(fan.set_fan_speed, 2)
        assert b
        assert fan.fan_level == 2

    def test_mode_toggle(self):
        """Test changing modes on air purifier."""
        self.mock_api.return_value = (orjson.dumps({'code': 0}), 200)
        fan = VeSyncAir131(DEV_LIST_DETAIL, self.manager)
        f = self.run_in_loop(fan.auto_mode)
        assert f
        assert fan.mode == 'auto'
        f = self.run_in_loop(fan.manual_mode)
        assert fan.mode == 'manual'
        assert f
        f = self.run_in_loop(fan.sleep_mode)
        assert fan.mode == 'sleep'
        assert f

    def test_airpur_set_timer(self):
        """Test timer function of Core*00S Purifiers."""
        self.mock_api.return_value = (
            orjson.dumps(call_json_fans.INNER_RESULT({'id': 1})), 200
            )
        fan = VeSyncAirBypass(CORE200S_DETAIL, self.manager)
        self.run_in_loop(fan.set_timer, 100)
        assert fan.timer is not None
        assert fan.timer.timer_duration == 100
        assert fan.timer.done is False
        assert fan.timer.action == 'off'
        assert fan.timer.running is True

    def test_airpur_clear_timer(self):
        """Test clear_timer method for Core air purifiers."""
        resp_dict, status = call_json_fans.FunctionResponses['Core200S']
        self.mock_api.return_value = orjson.dumps(resp_dict), status
        fan = VeSyncAirBypass(CORE200S_DETAIL, self.manager)
        fan.timer = call_json_fans.FAN_TIMER
        self.run_in_loop(fan.clear_timer)
        assert fan.timer is None
